#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import operator

try:
    import pandas as pd
    has_pandas = True
except ImportError:
    has_pandas = False

from ...compat import u, six, izip as zip, Version
from ...config import options
from ...console import get_console_size, in_interactive_session, \
    in_ipython_frontend, in_qtconsole
from ...utils import to_str, to_text, deprecated
from ...models import TableSchema
from ...types import Partition
from . import formatter as fmt


class ResultFrame(six.Iterator):
    class ResultRecord(list):
        def __init__(self, columns, values):
            self._columns = columns
            self._column_id_by_name = dict([(v.name, k) for k, v in enumerate(columns)])
            super(ResultFrame.ResultRecord, self).__init__(values)

        def __getitem__(self, item):
            if isinstance(item, six.string_types):
                item = self._column_id_by_name[item]
            return list.__getitem__(self, item)

        def iteritems(self):
            for col, val in zip(self._columns, self):
                yield col.name, val

        def keys(self):
            return [c.name for c in self._columns]

    def __init__(self, data, columns=None, schema=None, index=None, pandas=True):
        if columns is None and schema is None:
            raise ValueError('Either columns or schema should be provided')

        if columns is None and schema is not None:
            columns = schema.columns

        self._columns = columns
        self._names = [to_text(c.name) for c in self._columns]
        self._types = [c.type for c in self._columns]
        self._index = index

        if has_pandas and pandas:
            if isinstance(data, pd.DataFrame):
                self._values = self._reset_pd_axes(data)
            else:
                self._values = pd.DataFrame(
                    [self._get_values(r) for r in data], columns=self._names, index=index
                )
            self._index = self._values.index
            self._pandas = True
        else:
            if self._index is None:
                self._index = []
                self._values = []
                for i, r in zip(itertools.count(0), data):
                    self._values.append(self._get_values(r))
                    self._index.append(i)
            else:
                self._values = list(self._get_values(r) for r in data)
            self._pandas = False

        self._cursor = -1

    def _reset_pd_axes(self, data):
        if Version(pd.__version__) < Version("0.23.0"):
            data = data.values
            return pd.DataFrame(
                [self._get_values(r) for r in data], columns=self._names, index=self._index
            )

        ret_data = data
        if list(data.columns) != self._names:
            # already copied
            ret_data = ret_data.set_axis(self._names, axis="columns", inplace=False)

        if self._index is not None and list(data.index) != list(self._index):
            if data is ret_data:
                ret_data = ret_data.set_axis(self._index, axis="index", inplace=False)
            else:
                ret_data.set_axis(self._index, axis="index", inplace=True)
        elif self._index is None and (
            not isinstance(data.index, pd.RangeIndex)
            or data.index.start != 0
        ):
            if data is ret_data:
                ret_data = ret_data.reset_index(drop=True, inplace=False)
            else:
                ret_data.reset_index(drop=True, inplace=True)
        return ret_data

    def _get_values(self, r):
        if hasattr(r, 'values'):
            return r.values
        return r

    def __len__(self):
        return len(self.values)

    @property
    def columns(self):
        return self._columns

    @property
    def names(self):
        return self._names

    @property
    def types(self):
        return self._types

    @property
    def schema(self):
        return TableSchema(
            columns=[col for col in self._columns if not isinstance(col, Partition)],
            partitions=[col for col in self._columns if isinstance(col, Partition)],
        )

    @property
    def index(self):
        return self._index

    @property
    def values(self):
        return self._values

    def get_column_data(self, item):
        item = to_text(item)
        if item not in self.names:
            return None
        if self._pandas:
            if item in self.values.index.names:
                return self.values.index.get_level_values(item)
            else:
                return self.values[item]
        else:
            col_id = list(idx for idx, c in self.names if c == item)[0]
            return [r[col_id] for r in self.values]

    if has_pandas:
        def to_pandas(self, wrap=False):
            from .. import DataFrame
            if wrap:
                return DataFrame(self.values)
            else:
                return self.values

    def __getattr__(self, item):
        col = self.get_column_data(item)
        if col is None:
            raise AttributeError("ResultFrame has no column '{0}'".format(item))
        return col

    def __getitem__(self, item):
        if isinstance(item, six.integer_types):
            if self._pandas:
                return self._values.iloc[item]
            else:
                return self.ResultRecord(self._columns, self._values[item])
        elif isinstance(item, slice):
            if self._pandas:
                return self._values.iloc[item]
            else:
                return ResultFrame(self._values[item], columns=self._columns,
                                   index=self._index[item], pandas=self._pandas)
        elif isinstance(item, tuple) and len(item) == 2:
            if self._pandas:
                return self._values.iloc[item]
            else:
                if isinstance(item[1], slice):
                    frame = self[item[0]]
                    values = [r[item[1]] for r in frame._values]
                    return ResultFrame(values, columns=self._columns[item[1]],
                                       index=frame._index, pandas=self._pandas)
                else:
                    values = [r[item[1]] for r in self[item[0]]._values]
                    return values

    @deprecated("Direct call of next(ResultFrame) is deprecated")
    def __next__(self):
        self._cursor += 1
        try:
            if self._pandas:
                return self._values.iloc[self._cursor]
            else:
                return self.ResultRecord(self._columns, self._values[self._cursor])
        except IndexError:
            raise StopIteration

    def __iter__(self):
        cursor = 0
        while True:
            try:
                yield self[cursor]
                cursor += 1
            except IndexError:
                return

    def concat(self, *frames, **kwargs):
        if len(frames) == 2 and isinstance(frames[1], int):
            frames, axis = frames[:1], frames[1]
        else:
            axis = kwargs.pop("axis", 0)
        if kwargs:
            raise TypeError("Cannot accept arguments %s" % ",".join(kwargs))

        if len(frames) == 0:
            return self

        if self._pandas:
            try:
                from pandas import concat
            except ImportError:
                from pandas.tools.merge import concat

            values = (self._values,) + tuple(frame._values for frame in frames)
            pd_data = concat(values, axis=axis)
            if axis == 0:
                columns = self._columns
            else:
                columns = functools.reduce(
                    operator.add,
                    (self._columns,) + tuple(frame._columns for frame in frames),
                    )
            return ResultFrame(pd_data, columns=columns, pandas=True)
        else:
            if axis == 0:
                if any(self._columns != frame._columns for frame in frames):
                    raise ValueError(
                        'Cannot concat two frame of different columns')

                values = functools.reduce(
                    operator.add,
                    (self._values,) + tuple(frame._values for frame in frames),
                )
                indices = functools.reduce(
                    operator.add,
                    (self._index,) + tuple(frame._index for frame in frames),
                )
                return ResultFrame(values, columns=self._columns,
                                   index=indices, pandas=self._pandas)
            else:
                if any(self._index != frame._index for frame in frames):
                    raise ValueError(
                        'Cannot concat two frames of different indexes')

                sub_tuple = (self._values,) + tuple(frame._values for frame in frames)
                values = [
                    functools.reduce(operator.add, vals) for vals in zip(*sub_tuple)
                ]
                columns = functools.reduce(
                    operator.add,
                    (self._columns,) + tuple(frame._columns for frame in frames),
                )
                return ResultFrame(values, columns,
                                   index=self._index, pandas=self._pandas)

    @property
    def dtypes(self):
        return [it.type for it in self._columns]

    def __len__(self):
        return len(self._values)

    def _repr_fits_vertical_(self):
        """
        Check length against max_rows.
        """
        max_rows = options.display.max_rows
        return len(self) <= max_rows

    def _repr_fits_horizontal_(self, ignore_width=False):
        """
        Check if full repr fits in horizontal boundaries imposed by the display
        options width and max_columns. In case off non-interactive session, no
        boundaries apply.

        ignore_width is here so ipnb+HTML output can behave the way
        users expect. display.max_columns remains in effect.
        GH3541, GH3573
        """

        width, height = get_console_size()
        max_columns = options.display.max_columns
        nb_columns = len(self.columns)

        # exceed max columns
        if ((max_columns and nb_columns > max_columns) or
                ((not ignore_width) and width and nb_columns > (width // 2))):
            return False

        if (ignore_width  # used by repr_html under IPython notebook
                # scripts ignore terminal dims
                or not in_interactive_session()):
            return True

        if (options.display.width is not None or
                in_ipython_frontend()):
            # check at least the column row for excessive width
            max_rows = 1
        else:
            max_rows = options.display.max_rows

        # when auto-detecting, so width=None and not in ipython front end
        # check whether repr fits horizontal by actualy checking
        # the width of the rendered repr
        buf = six.StringIO()

        # only care about the stuff we'll actually print out
        # and to_string on entire frame may be expensive
        d = self

        if not (max_rows is None):  # unlimited rows
            # min of two, where one may be None
            d = d[:min(max_rows, len(d))]
        else:
            return True

        d.to_string(buf=buf)
        value = buf.getvalue()
        repr_width = max([len(l) for l in value.split('\n')])

        return repr_width < width

    def __repr__(self):
        return to_str(self.__unicode__())

    def __unicode__(self):
        """
        Return a string representation for a particular DataFrame
        """
        if self._pandas:
            return to_text(repr(self._values))

        buf = six.StringIO(u(""))

        max_rows = options.display.max_rows
        max_cols = options.display.max_columns
        show_dimensions = options.display.show_dimensions
        if options.display.expand_frame_repr:
            width, _ = get_console_size()
        else:
            width = None
        self.to_string(buf=buf, max_rows=max_rows, max_cols=max_cols,
                       line_width=width, show_dimensions=show_dimensions)

        return to_str(buf.getvalue())

    def _repr_html_(self):
        """
        Return a html representation for a particular DataFrame.
        Mainly for IPython notebook.
        """
        # qtconsole doesn't report its line width, and also
        # behaves badly when outputting an HTML table
        # that doesn't fit the window, so disable it.
        # XXX: In IPython 3.x and above, the Qt console will not attempt to
        # display HTML, so this check can be removed when support for IPython 2.x
        # is no longer needed.

        if self._pandas and options.display.notebook_widget:
            from .. import DataFrame
            from ..ui import show_df_widget
            show_df_widget(DataFrame(self._values, schema=self.schema))

        if self._pandas:
            return self._values._repr_html_()

        if in_qtconsole():
            # 'HTML output is disabled in QtConsole'
            return None

        if options.display.notebook_repr_html:
            max_rows = options.display.max_rows
            max_cols = options.display.max_columns
            show_dimensions = options.display.show_dimensions

            return self.to_html(max_rows=max_rows, max_cols=max_cols,
                                show_dimensions=show_dimensions,
                                notebook=True)
        else:
            return None

    def to_string(self, buf=None, columns=None, col_space=None,
                  header=True, index=True, na_rep='NaN', formatters=None,
                  float_format=None, sparsify=None, index_names=True,
                  justify=None, line_width=None, max_rows=None, max_cols=None,
                  show_dimensions=False):
        """
        Render a DataFrame to a console-friendly tabular output.
        """

        formatter = fmt.ResultFrameFormatter(self, buf=buf, columns=columns,
                                             col_space=col_space, na_rep=na_rep,
                                             formatters=formatters,
                                             float_format=float_format,
                                             sparsify=sparsify,
                                             justify=justify,
                                             index_names=index_names,
                                             header=header, index=index,
                                             line_width=line_width,
                                             max_rows=max_rows,
                                             max_cols=max_cols,
                                             show_dimensions=show_dimensions)
        formatter.to_string()

        if buf is None:
            result = formatter.buf.getvalue()
            return result

    def to_html(self, buf=None, columns=None, col_space=None,
                header=True, index=True, na_rep='NaN', formatters=None,
                float_format=None, sparsify=None, index_names=True,
                justify=None, bold_rows=True, classes=None, escape=True,
                max_rows=None, max_cols=None, show_dimensions=False,
                notebook=False):
        """
        Render a DataFrame as an HTML table.

        `to_html`-specific options:

        bold_rows : boolean, default True
            Make the row labels bold in the output
        classes : str or list or tuple, default None
            CSS class(es) to apply to the resulting html table
        escape : boolean, default True
            Convert the characters <, >, and & to HTML-safe sequences.=
        max_rows : int, optional
            Maximum number of rows to show before truncating. If None, show
            all.
        max_cols : int, optional
            Maximum number of columns to show before truncating. If None, show
            all.

        """

        formatter = fmt.ResultFrameFormatter(self, buf=buf, columns=columns,
                                             col_space=col_space, na_rep=na_rep,
                                             formatters=formatters,
                                             float_format=float_format,
                                             sparsify=sparsify,
                                             justify=justify,
                                             index_names=index_names,
                                             header=header, index=index,
                                             bold_rows=bold_rows,
                                             escape=escape,
                                             max_rows=max_rows,
                                             max_cols=max_cols,
                                             show_dimensions=show_dimensions)
        formatter.to_html(classes=classes, notebook=notebook)

        if buf is None:
            return formatter.buf.getvalue()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..analyzer import BaseAnalyzer
from ...expr.element import Cut, Between
from ...expr.arithmetic import *
from ...expr.math import Log1p, Expm1
from ...expr.strings import *
from ...expr.window import CumSum
from ...expr.reduction import *
from ...expr.collections import PivotCollectionExpr
from ... import types


class Analyzer(BaseAnalyzer):
    def visit_element_op(self, expr):
        if isinstance(expr, Between) and not expr.inclusive:
            sub = ((expr.left < expr.input) & (expr.input.copy() < expr.right))
            self._sub(expr, sub.rename(expr.name))
        elif isinstance(expr, Cut):
            sub = self._get_cut_sub_expr(expr)
            self._sub(expr, sub)

    def visit_value_counts(self, expr):
        self._sub(expr, self._get_value_counts_sub_expr(expr))

    def visit_binary_op(self, expr):
        if isinstance(expr, Divide) and isinstance(expr._lhs.dtype, types.Integer) and \
                isinstance(expr._rhs.dtype, types.Integer):
            self._sub(expr._lhs, expr._lhs.astype('float'), parents=[expr])
        elif isinstance(expr, FloorDivide):
            lhs = (expr._lhs < 0).ifelse(expr._lhs - 1, expr._lhs)
            self._sub(expr._lhs, lhs, parents=[expr])

    def visit_unary_op(self, expr):
        if isinstance(expr, Invert) and isinstance(expr.input.dtype, types.Integer):
            sub = -expr._input - 1
            self._sub(expr, sub)
        elif isinstance(expr, Negate) and expr.input.dtype == types.boolean:
            input = expr._input.ifelse(1, 0)
            self._sub(expr.input, input)

    def visit_math(self, expr):
        if isinstance(expr, Log1p):
            sub = (1 + expr.input).log()
            self._sub(expr, sub)
        elif isinstance(expr, Expm1):
            sub = expr.input.exp() - 1
            self._sub(expr, sub)

    def visit_string_op(self, expr):
        if isinstance(expr, Ljust):
            rest = expr.width - expr.input.len()
            sub = expr.input + \
                     (rest >= 0).ifelse(expr._fillchar.repeat(rest), '')
            self._sub(expr, sub.rename(expr.name))
        elif isinstance(expr, Rjust):
            rest = expr.width - expr.input.len()
            sub = (rest >= 0).ifelse(expr._fillchar.repeat(rest), '') + expr.input
            self._sub(expr, sub.rename(expr.name))
        elif isinstance(expr, Zfill):
            fillchar = Scalar('0')
            rest = expr.width - expr.input.len()
            sub = (rest >= 0).ifelse(fillchar.repeat(rest), '') + expr.input
            self._sub(expr, sub.rename(expr.name))
        elif isinstance(expr, CatStr):
            input = expr.input
            others = expr._others if isinstance(expr._others, Iterable) else (expr._others, )
            for other in others:
                if expr.na_rep is not None:
                    for e in (input, ) + tuple(others):
                        self._sub(e, e.fillna(expr.na_rep), parents=(expr, ))
                    return
                else:
                    if expr._sep is not None:
                        input = other.isnull().ifelse(input, input + expr._sep + other)
                    else:
                        input = other.isnull().ifelse(input, input + other)
            self._sub(expr, input.rename(expr.name))

    def visit_cum_window(self, expr):
        if isinstance(expr, CumSum) and expr.input.dtype == types.boolean:
            sub = expr.input.ifelse(1, 0)
            self._sub(expr.input, sub, parents=[expr])

    def _visit_pivot(self, expr):
        sub = self._get_pivot_sub_expr(expr)
        self._sub(expr, sub)

    def _visit_pivot_table(self, expr):
        sub = self._get_pivot_table_sub_expr(expr)
        self._sub(expr, sub)

    def visit_pivot(self, expr):
        if isinstance(expr, PivotCollectionExpr):
            self._visit_pivot(expr)
        else:
            self._visit_pivot_table(expr)

    def visit_reduction(self, expr):
        if isinstance(expr, (Max, GroupedMax, Min, GroupedMin, Sum, GroupedSum)) and \
                expr.input.dtype == types.boolean:
            self._sub(expr.input, expr.input.ifelse(1, 0), parents=[expr])

        if isinstance(expr, (Sum, GroupedSum)) and expr.input.dtype == types.string:
            sub = expr.input.cat(sep='')
            if isinstance(expr, GroupedSum):
                sub = sub.to_grouped_reduction(expr._grouped)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (Any, GroupedAny)):
            input = expr.input.ifelse(1, 0).max() == 1
            if isinstance(expr, GroupedAny):
                input = input.to_grouped_reduction(expr._grouped)
            sub = input.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (All, GroupedAll)):
            input = expr.input.ifelse(1, 0).min() == 1
            if isinstance(expr, GroupedAll):
                input = input.to_grouped_reduction(expr._grouped)
            sub = input.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (Moment, GroupedMoment)):
            order = expr._order
            center = expr._center

            sub = self._get_moment_sub_expr(expr, expr.input, order, center)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (Skewness, GroupedSkewness)):
            std = expr.input.std(ddof=1)
            if isinstance(expr, GroupedSequenceReduction):
                std = std.to_grouped_reduction(expr._grouped)
            cnt = expr.input.count()
            if isinstance(expr, GroupedSequenceReduction):
                cnt = cnt.to_grouped_reduction(expr._grouped)
            sub = self._get_moment_sub_expr(expr, expr.input, 3, True) / (std ** 3)
            sub *= (cnt ** 2) / (cnt - 1) / (cnt - 2)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)
        elif isinstance(expr, (Kurtosis, GroupedKurtosis)):
            std = expr.input.std(ddof=0)
            if isinstance(expr, GroupedSequenceReduction):
                std = std.to_grouped_reduction(expr._grouped)
            cnt = expr.input.count()
            if isinstance(expr, GroupedSequenceReduction):
                cnt = cnt.to_grouped_reduction(expr._grouped)
            m4 = self._get_moment_sub_expr(expr, expr.input, 4, True)
            sub = 1.0 / (cnt - 2) / (cnt - 3) * ((cnt * cnt - 1) * m4 / (std ** 4) - 3 * (cnt - 1) ** 2)
            sub = sub.rename(expr.name)
            self._sub(expr, sub)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from .. import serializers
from ..compat import six
from ..errors import InternalServerError, InvalidParameter, MethodNotAllowed, NoSuchObject
from ..utils import with_wait_argument
from .core import Iterable
from .schema import Schema

logger = logging.getLogger(__name__)
_project_has_schema_api = dict()


def with_schema_api_fallback(fallback_fun, is_iter=False):
    def decorator(fun):
        @six.wraps(fun)
        def wrapper(self, *args, **kwargs):
            key = (self.parent.odps.endpoint, self.parent.name)
            kw = kwargs.copy()
            try:
                self._check_schema_api()
                result = fun(self, *args, **kw)
                _project_has_schema_api[key] = True
                return result
            except (MethodNotAllowed, InvalidParameter):
                _project_has_schema_api[key] = False

            return fallback_fun(self, *args, **kwargs)

        @six.wraps(fun)
        def iter_wrapper(self, *args, **kwargs):
            key = (self.parent.odps.endpoint, self.parent.name)
            kw = kwargs.copy()
            try:
                self._check_schema_api()
                for item in fun(self, *args, **kw):
                    yield item
                    _project_has_schema_api[key] = True
                return
            except (MethodNotAllowed, InvalidParameter):
                if _project_has_schema_api.get(key):
                    # in case duplicated items are iterated
                    raise
                _project_has_schema_api[key] = False

            for item in fallback_fun(self, *args, **kwargs):
                yield item

        return iter_wrapper if is_iter else wrapper
    return decorator


class Schemas(Iterable):
    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems')
    schemas = serializers.XMLNodesReferencesField(Schema, 'Schema')

    def __iter__(self):
        return self.iterate()

    def resource(self, client=None, endpoint=None):
        return self.parent.resource(client, endpoint=endpoint)

    def _check_schema_api(self):
        key = (self.parent.odps.endpoint, self.parent.name)
        if not _project_has_schema_api.get(key, True):
            raise MethodNotAllowed("Schema API not supported")

    def _iterate_legacy(self, name=None, owner=None):
        if name is not None or owner is not None:
            raise ValueError(
                "Iterating schemas with name or owner not supported on current service"
            )
        inst = self.parent.odps.execute_sql("SHOW SCHEMAS IN %s" % self.parent.name)
        schema_names = inst.get_task_results().get("AnonymousSQLTask").strip().split("\n")
        for schema_name in schema_names:
            yield Schema(name=schema_name, parent=self, client=self._client)

    @with_schema_api_fallback(fallback_fun=_iterate_legacy, is_iter=True)
    def iterate(self, name=None, owner=None):
        params = {'expectmarker': 'true'}
        if name is not None:
            params['name'] = name
        if owner is not None:
            params['owner'] = owner
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource() + "/schemas"
            resp = self._client.get(url, params=params)

            r = Schemas.parse(self._client, resp, obj=self)
            params['marker'] = r.marker

            return r.schemas

        while True:
            schemas = _it()
            if schemas is None:
                break
            for schema in schemas:
                yield schema

    @with_wait_argument
    def _create_legacy(self, obj=None, async_=False, **kwargs):
        schema_name = kwargs.pop("schema_name", obj)
        if isinstance(obj, Schema):
            schema_name = obj.name
        inst = self.parent.odps.run_sql(
            "CREATE SCHEMA %s.%s" % (self.parent.name, schema_name)
        )
        if not async_:
            inst.wait_for_success()
            return Schema(name=schema_name, parent=self, client=self._client)
        return inst

    @with_schema_api_fallback(fallback_fun=_create_legacy)
    def create(self, obj=None, **kwargs):
        kwargs.pop("async_", None)
        kwargs.pop("wait", None)

        if isinstance(obj, six.string_types):
            kwargs["name"] = obj
            obj = None
        schema = obj or Schema(parent=self, client=self._client, **kwargs)

        if schema.parent is None:
            schema._parent = self
        if schema._client is None:
            schema._client = self._client

        headers = {'Content-Type': 'application/xml'}
        data = schema.serialize()

        resource = self.resource() + "/schemas"
        self._client.post(resource, data, headers=headers)
        return schema

    @with_wait_argument
    def _delete_legacy(self, schema_name, async_=False):
        if isinstance(schema_name, Schema):
            schema_name = schema_name.name
        inst = self.parent.odps.run_sql(
            "DROP SCHEMA %s.%s" % (self.parent.name, schema_name)
        )
        if not async_:
            return inst.wait_for_success()
        return inst

    @with_schema_api_fallback(fallback_fun=_delete_legacy)
    def delete(self, schema_name, async_=False):
        if isinstance(schema_name, Schema):
            schema_name = schema_name.name
        resource = self.resource() + "/schemas/" + schema_name
        self._client.delete(resource)

    def _get(self, item):
        if isinstance(item, Schema):
            return item
        return Schema(name=item, parent=self, client=self._client)

    def _contains_legacy(self, item):
        try:
            next(self._get(item).functions)
        except StopIteration:
            pass
        except NoSuchObject:
            return False
        except InvalidParameter as ex:
            if "NoSuchObjectException" in str(ex):
                return False
            raise
        except InternalServerError as ex:
            if "invalid schema name" in str(ex).lower():
                return False
            raise

        return True

    @with_schema_api_fallback(fallback_fun=_contains_legacy)
    def __contains__(self, item):
        schema = self._get(item)
        try:
            schema.reload()
            return True
        except NoSuchObject:
            return False



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: kv_store_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='kv_store_service.proto',
  package='apsara.odps.cupid.protocol',
  syntax='proto2',
  serialized_pb=_b('\n\x16kv_store_service.proto\x12\x1a\x61psara.odps.cupid.protocol\"(\n\nPutRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\"\r\n\x0bPutResponse\"\x1b\n\nGetRequest\x12\r\n\x05value\x18\x01 \x01(\t\"\x1c\n\x0bGetResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x32\xc0\x01\n\x0eKVStoreService\x12V\n\x03Put\x12&.apsara.odps.cupid.protocol.PutRequest\x1a\'.apsara.odps.cupid.protocol.PutResponse\x12V\n\x03Get\x12&.apsara.odps.cupid.protocol.GetRequest\x1a\'.apsara.odps.cupid.protocol.GetResponseB\t\x80\x01\x01\x88\x01\x01\x90\x01\x01')
)




_PUTREQUEST = _descriptor.Descriptor(
  name='PutRequest',
  full_name='apsara.odps.cupid.protocol.PutRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='apsara.odps.cupid.protocol.PutRequest.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='apsara.odps.cupid.protocol.PutRequest.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=54,
  serialized_end=94,
)


_PUTRESPONSE = _descriptor.Descriptor(
  name='PutResponse',
  full_name='apsara.odps.cupid.protocol.PutResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=96,
  serialized_end=109,
)


_GETREQUEST = _descriptor.Descriptor(
  name='GetRequest',
  full_name='apsara.odps.cupid.protocol.GetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='apsara.odps.cupid.protocol.GetRequest.value', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=111,
  serialized_end=138,
)


_GETRESPONSE = _descriptor.Descriptor(
  name='GetResponse',
  full_name='apsara.odps.cupid.protocol.GetResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='apsara.odps.cupid.protocol.GetResponse.value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=140,
  serialized_end=168,
)

DESCRIPTOR.message_types_by_name['PutRequest'] = _PUTREQUEST
DESCRIPTOR.message_types_by_name['PutResponse'] = _PUTRESPONSE
DESCRIPTOR.message_types_by_name['GetRequest'] = _GETREQUEST
DESCRIPTOR.message_types_by_name['GetResponse'] = _GETRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PutRequest = _reflection.GeneratedProtocolMessageType('PutRequest', (_message.Message,), dict(
  DESCRIPTOR = _PUTREQUEST,
  __module__ = 'kv_store_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.PutRequest)
  ))
_sym_db.RegisterMessage(PutRequest)

PutResponse = _reflection.GeneratedProtocolMessageType('PutResponse', (_message.Message,), dict(
  DESCRIPTOR = _PUTRESPONSE,
  __module__ = 'kv_store_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.PutResponse)
  ))
_sym_db.RegisterMessage(PutResponse)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETREQUEST,
  __module__ = 'kv_store_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetRequest)
  ))
_sym_db.RegisterMessage(GetRequest)

GetResponse = _reflection.GeneratedProtocolMessageType('GetResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETRESPONSE,
  __module__ = 'kv_store_service_pb2'
  # @@protoc_insertion_point(class_scope:apsara.odps.cupid.protocol.GetResponse)
  ))
_sym_db.RegisterMessage(GetResponse)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('\200\001\001\210\001\001\220\001\001'))

_KVSTORESERVICE = _descriptor.ServiceDescriptor(
  name='KVStoreService',
  full_name='apsara.odps.cupid.protocol.KVStoreService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=171,
  serialized_end=363,
  methods=[
  _descriptor.MethodDescriptor(
    name='Put',
    full_name='apsara.odps.cupid.protocol.KVStoreService.Put',
    index=0,
    containing_service=None,
    input_type=_PUTREQUEST,
    output_type=_PUTRESPONSE,
    options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='apsara.odps.cupid.protocol.KVStoreService.Get',
    index=1,
    containing_service=None,
    input_type=_GETREQUEST,
    output_type=_GETRESPONSE,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_KVSTORESERVICE)

DESCRIPTOR.services_by_name['KVStoreService'] = _KVSTORESERVICE

KVStoreService = service_reflection.GeneratedServiceType('KVStoreService', (_service.Service,), dict(
  DESCRIPTOR = _KVSTORESERVICE,
  __module__ = 'kv_store_service_pb2'
  ))

KVStoreService_Stub = service_reflection.GeneratedServiceStubType('KVStoreService_Stub', (KVStoreService,), dict(
  DESCRIPTOR = _KVSTORESERVICE,
  __module__ = 'kv_store_service_pb2'
  ))


# @@protoc_insertion_point(module_scope)



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import groupby

from ..enums import PortType, PortDirection
from ...compat import six
from ...utils import str_to_bool, camel_to_underline
from ... import types as odps_types, serializers


class ParamDef(serializers.XMLSerializableModel):
    """
    Description for algorithm parameter.

    :param name: parameter name
    :type name: str
    :param default: default value
    :type default: str
    :param type: parameter type
    :type type: str
    """
    _root = 'param'

    name = serializers.XMLNodeAttributeField(attr='name')
    required = serializers.XMLNodeAttributeField(attr='required', parse_callback=str_to_bool)
    docs = serializers.XMLNodeField('docs')
    alias = serializers.XMLNodeField('alias')
    _value = serializers.XMLNodeField('value')
    type = serializers.XMLNodeField('type')
    exported = serializers.XMLNodeField('exported', parse_callback=str_to_bool)
    internal = serializers.XMLNodeField('internal', parse_callback=str_to_bool)
    exporter = serializers.XMLNodeField('exporter')
    setter = serializers.XMLNodeField('setter')
    input_name = serializers.XMLNodeField('inputName')
    output_name = serializers.XMLNodeField('outputName')
    sequence = serializers.XMLNodeField('sequence', parse_callback=int)

    def __init__(self, *args, **kwargs):
        if len(args) >= 1 and 'name' not in kwargs:
            kwargs['name'] = args[0]
        if len(args) >= 2 and 'default' not in kwargs:
            kwargs['value'] = args[1]
        super(ParamDef, self).__init__(**kwargs)
        self.exported = True if self.exported is None else self.exported
        if kwargs.get('default') is not None:
            self._value = kwargs.get('default')

    @property
    def friendly_name(self):
        return camel_to_underline(self.name) if not self.alias else camel_to_underline(self.alias)

    @property
    def value(self):
        if self._value is None or self.type is None:
            return self._value
        odps_type = getattr(odps_types, self.type)
        return odps_type.cast_value(self._value, odps_types.infer_primitive_data_type(self._value))

    @value.setter
    def value(self, v):
        self._value = v

    def copy(self):
        cls = self.__class__
        keys = dict((k, getattr(self, k)) for k in six.iterkeys(getattr(cls, '__fields')))
        return cls(**keys)

    @classmethod
    def build_input_table(cls, name='inputTableName', input_name='input'):
        """
        Build an input table parameter

        :param name: parameter name
        :type name: str
        :param input_name: bind input port name
        :param input_name: str
        :return: input description
        :rtype: ParamDef
        """
        obj = cls(name)
        obj.exporter = 'get_input_table_name'
        obj.input_name = input_name
        return obj

    @classmethod
    def build_input_partitions(cls, name='inputTablePartitions', input_name='input'):
        """
        Build an input table partition parameter

        :param name: parameter name
        :type name: str
        :param input_name: bind input port name
        :param input_name: str
        :return: input description
        :rtype: ParamDef
        """
        obj = cls(name)
        obj.exporter = 'get_input_partitions'
        obj.input_name = input_name
        return obj

    @classmethod
    def build_feature_col_names(cls, name='featureColNames', input_name='input'):
        obj = cls(name)
        obj.exporter = 'get_feature_columns'
        obj.input_name = input_name
        return obj

    @classmethod
    def build_label_col_name(cls, name='labelColName', input_name='input'):
        obj = cls(name)
        obj.exporter = 'get_label_column'
        obj.input_name = input_name
        return obj

    @classmethod
    def build_output_table(cls, name='inputTableName', output_name='output'):
        """
        Build an output table parameter

        :param name: parameter name
        :type name: str
        :param output_name: bind input port name
        :type output_name: str
        :return: output description
        :rtype: ParamDef
        """
        obj = cls(name)
        obj.exporter = 'get_output_table_name'
        obj.output_name = output_name
        return obj

    @classmethod
    def build_output_partitions(cls, name='inputTablePartitions', output_name='output'):
        """
        Build an output table partition parameter

        :param name: parameter name
        :type name: str
        :param output_name: bind input port name
        :type output_name: str
        :return: output description
        :rtype: ParamDef
        """
        obj = cls(name)
        obj.exporter = 'get_output_table_partition'
        obj.output_name = output_name
        return obj

    @classmethod
    def build_model_name(cls, name='modelName', output_name='output'):
        """
        Build an output model name parameter.

        :param name: model name
        :type name: str
        :param output_name: bind output port name
        :type output_name: str
        :return: output description
        :rtype: ParamDef
        """
        obj = cls(name)
        obj.exporter = 'generate_model_name'
        obj.output_name = output_name
        return obj


class SchemaDef(serializers.XMLSerializableModel):
    name = serializers.XMLNodeAttributeField(attr='name')
    dynamic = serializers.XMLNodeField('dynamic', type='bool')
    copy_input = serializers.XMLNodeField('copyInput')
    schema = serializers.XMLNodeField('schema')
    direct_copy = serializers.XMLNodeField('directCopy')
    mixin = serializers.XMLNodeField('mixin')

    def __init__(self, **kwargs):
        super(SchemaDef, self).__init__(**kwargs)
        self.dynamic = False if self.dynamic is None else self.dynamic

    @property
    def programmatic(self):
        return self.schema and '.' in self.schema


class ModelDef(serializers.XMLSerializableModel):
    type = serializers.XMLNodeField('type', default='PmmlModel')
    schemas = serializers.XMLNodesReferencesField(SchemaDef, 'schemas', 'schema')
    copy_params = serializers.XMLNodeField('copyParams')
    mixin = serializers.XMLNodeField('mixin')


class PortDef(serializers.XMLSerializableModel):
    """
    Description for algorithm port.

    :param name: port name
    :type name: str
    :param direction: port direction (input / output)
    :type direction: PortDirection
    :param type: port type (model / data)
    :type type: PortType
    :param copy_input: input name where the schema is copied from.
    :type copy_input: str
    :param schema: k1:v1,k2:v2 string describing the schema to be appended
    :type schema: str
    """
    name = serializers.XMLNodeAttributeField(attr='name')
    docs = serializers.XMLNodeField('docs')
    type = serializers.XMLNodeField('type', parse_callback=lambda v: PortType(v.upper()),
                                    serialize_callback=lambda v: v.value)
    io_type = serializers.XMLNodeField('ioType', parse_callback=lambda v: PortDirection(v.upper()),
                                       serialize_callback=lambda v: v.value)
    sequence = serializers.XMLNodeField('sequence', parse_callback=int)
    required = serializers.XMLNodeField('required', parse_callback=str_to_bool, default=True)
    schema = serializers.XMLNodeReferenceField(SchemaDef, 'schema')
    model = serializers.XMLNodeReferenceField(ModelDef, 'model')
    exporter = serializers.XMLNodeField('exporter')

    def __init__(self, *args, **kwargs):
        if len(args) >= 1 and 'name' not in kwargs:
            kwargs['name'] = args[0]
        if len(args) >= 2 and 'io_type' not in kwargs:
            kwargs['io_type'] = args[1]
        super(PortDef, self).__init__(**kwargs)

    @classmethod
    def build_data_input(cls, name='input'):
        """
        Build a data input port.

        :param name: port name
        :type name: str
        :return: port object
        :rtype: PortDef
        """
        return cls(name, PortDirection.INPUT, type=PortType.DATA)

    @classmethod
    def build_data_output(cls, name='output', copy_input=None, schema=None):
        """
        Build a data output port.

        :param name: port name
        :type name: str
        :return: port object
        :param copy_input: input name where the schema is copied from.
        :type copy_input: str
        :param schema: k1:v1,k2:v2 string describing the schema to be appended
        :type schema: str
        :rtype: PortDef
        """
        return cls(name, PortDirection.OUTPUT, type=PortType.DATA, copy_input=copy_input, schema=schema)

    @classmethod
    def build_model_input(cls, name='input'):
        """
        Build a model input port.

        :param name: port name
        :type name: str
        :return: port object
        :rtype: PortDef
        """
        return cls(name, PortDirection.INPUT, type=PortType.MODEL)

    @classmethod
    def build_model_output(cls, name='output'):
        """
        Build a model output port.

        :param name: port name
        :type name: str
        :return: port object
        :rtype: PortDef
        """
        return cls(name, PortDirection.OUTPUT, type=PortType.MODEL)


class MetaDef(serializers.XMLSerializableModel):
    name = serializers.XMLNodeAttributeField(attr='name')
    _param_value = serializers.XMLNodeAttributeField(attr='value')
    _text_value = serializers.XMLNodeField('.')

    @property
    def value(self):
        return self._text_value or self._param_value

    @value.setter
    def value(self, v):
        self._text_value = v
        self._param_value = None

    def __init__(self, name=None, value=None, **kwargs):
        if name:
            kwargs['name'] = name
        if value:
            kwargs['_text_value'] = value
        super(MetaDef, self).__init__(**kwargs)


class AlgorithmDef(serializers.XMLSerializableModel):
    """
    Description for an algorithm. This class can help users publish their own algorithms.

    :param code_name: Code name of the algorithm

    :type params: list[ParamDef]
    :type ports: list[PortDef]
    :type metas: list[MetaDef]
    :type code_name: str
    """
    _root = 'algorithm'
    code_name = serializers.XMLNodeAttributeField(attr='codeName')
    docs = serializers.XMLNodeField('docs')
    entry_docs = serializers.XMLNodeField('entryDocs')
    export_function = serializers.XMLNodeField('exportFunction', parse_callback=str_to_bool)
    enabled = serializers.XMLNodeField('enabled', parse_callback=str_to_bool)
    reload_fields = serializers.XMLNodeField('reloadFields', parse_callback=str_to_bool)
    base_class = serializers.XMLNodeField('baseClass')
    params = serializers.XMLNodesReferencesField(ParamDef, 'params', 'param')
    ports = serializers.XMLNodesReferencesField(PortDef, 'ports', 'port')
    metas = serializers.XMLNodesReferencesField(MetaDef, 'metas', 'meta')
    public = serializers.XMLNodeField('public', parse_callback=str_to_bool, default=True)
    port_seqs = dict()

    def __init__(self, code_name=None, **kwargs):
        super(AlgorithmDef, self).__init__(**kwargs)
        self.code_name = code_name
        self.params = []
        self.ports = []
        self.metas = []

    def add_param(self, param):
        """
        Add a parameter object to the definition

        :param param: parameter
        :type param: ParamDef
        """
        self.params.append(param)
        return self

    def add_port(self, port):
        """
        Add a port object to the definition

        :param port: port definition
        :type port: PortDef
        """
        self.ports.append(port)
        if port.io_type not in self.port_seqs:
            self.port_seqs[port.io_type] = 0
        self.port_seqs[port.io_type] += 1
        port.sequence = self.port_seqs[port.io_type]
        return self

    def add_meta(self, name, value):
        """
        Add a pair of meta data to the definition

        :param name: name of the meta
        :type name: str
        :param value: value of the meta
        :type value: str
        """
        for mt in self.metas:
            if mt.name == name:
                mt.value = value
                return self
        self.metas.append(MetaDef(name, value))
        return self

    def get_input_ports(self):
        return sorted([p for p in self.ports if p.io_type == PortDirection.INPUT], key=lambda p: p.sequence)

    def get_output_ports(self):
        return sorted([p for p in self.ports if p.io_type == PortDirection.OUTPUT], key=lambda p: p.sequence)

    @property
    def meta_dict(self):
        return dict((m.name, m.value) for m in self.metas)

    def serialize(self):
        """
        Serialize the algorithm definition
        """
        # fill sequences
        for keys, groups in groupby(self.ports, lambda x: x.io_type):
            for seq, port in enumerate(groups):
                port.sequence = seq
        return super(AlgorithmDef, self).serialize()


class XflowAlgorithmDef(AlgorithmDef):
    """
    Description for Xflow algorithm.
    """

    def __init__(self, code_name=None, xflow_name=None, project='algo_public', **kwargs):
        super(XflowAlgorithmDef, self).__init__(code_name, **kwargs)
        if xflow_name is None:
            xflow_name = code_name
        if xflow_name is not None:
            self.add_meta('xflowName', xflow_name)
        self.add_meta('xflowProjectName', project)


class AlgorithmsDef(serializers.XMLSerializableModel):
    algorithms = serializers.XMLNodesReferencesField(AlgorithmDef, 'algorithm')
    base_class = serializers.XMLNodeAttributeField(attr='baseClass')

    @classmethod
    def parse(cls, response, obj=None, **kw):
        if six.PY2:
            response = response.encode('utf-8')
        return super(AlgorithmsDef, cls).parse(response, obj=obj, **kw)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .element import ElementWise
from .expressions import Expr, StringSequenceExpr, Scalar, StringScalar, SequenceExpr
from . import utils
from .. import types
from ...compat import six


class StringOp(ElementWise):
    __slots__ = ()

    def _init(self, *args, **kwargs):
        for arg in self._args[1:]:
            self._init_attr(arg, None)

        super(StringOp, self)._init(*args, **kwargs)

        for attr in self._args[1:]:
            val = getattr(self, attr)
            if val is not None and not isinstance(val, (Expr, list, tuple)):
                setattr(self, attr, Scalar(_value=val))

    def __getattribute__(self, attr):
        if attr in ('input', '_input'):
            return super(StringOp, self).__getattribute__(attr)
        else:
            try:
                return object.__getattribute__(self, attr)
            except AttributeError as e:
                err = e
            if not attr.startswith('_'):
                private_attr = '_%s' % attr
                try:
                    scalar = object.__getattribute__(self, private_attr)
                    if isinstance(scalar, Scalar):
                        return scalar.value
                    return scalar
                except AttributeError:
                    raise err

    def accept(self, visitor):
        return visitor.visit_string_op(self)


class Capitalize(StringOp):
    __slots__ = ()


class CatStr(StringOp):
    _args = '_input', '_others', '_sep', '_na_rep'
    _add_args_slots = False

    @property
    def node_name(self):
        return 'Cat'


class Contains(StringOp):
    _args = '_input', '_pat', '_case', '_flags', '_regex'
    _add_args_slots = False


class Count(StringOp):
    _args = '_input', '_pat', '_flags'
    _add_args_slots = False


class Endswith(StringOp):
    _args = '_input', '_pat'
    _add_args_slots = False


class Startswith(StringOp):
    _args = '_input', '_pat'
    _add_args_slots = False


class Extract(StringOp):
    _args = '_input', '_pat', '_flags', '_group'
    _add_args_slots = False


class Find(StringOp):
    _args = '_input', '_sub', '_start', '_end'
    _add_args_slots = False


class RFind(StringOp):
    _args = '_input', '_sub', '_start', '_end'
    _add_args_slots = False


class Replace(StringOp):
    _args = '_input', '_pat', '_repl', '_n', '_case', '_flags', '_regex'
    _add_args_slots = False


class Get(StringOp):
    _args = '_input', '_index'
    _add_args_slots = False


class Join(StringOp):
    _args = '_input', '_sep'
    _add_args_slots = False


class Len(StringOp):
    _args = '_input',
    _add_args_slots = False


class Ljust(StringOp):
    _args = '_input', '_width', '_fillchar'
    _add_args_slots = False


class Rjust(StringOp):
    _args = '_input', '_width', '_fillchar'
    _add_args_slots = False


class Lower(StringOp):
    _args = '_input',
    _add_args_slots = False


class Upper(StringOp):
    _args = '_input',
    _add_args_slots = False


class Lstrip(StringOp):
    _args = '_input', '_to_strip'
    _add_args_slots = False


class Rstrip(StringOp):
    _args = '_input', '_to_strip'
    _add_args_slots = False


class Strip(StringOp):
    _args = '_input', '_to_strip'
    _add_args_slots = False


class Pad(StringOp):
    _args = '_input', '_width', '_side', '_fillchar'
    _add_args_slots = False

    def _init(self, *args, **kwargs):
        super(Pad, self)._init(*args, **kwargs)

        if self.side not in ('left', 'right', 'both'):
            raise ValueError('Side should be left, right or both')


class Repeat(StringOp):
    _args = '_input', '_repeats'
    _add_args_slots = False


class Split(StringOp):
    _args = '_input', '_pat', '_n'
    _add_args_slots = False


class RSplit(StringOp):
    _args = '_input', '_pat', '_n'
    _add_args_slots = False


class Slice(StringOp):
    _args = '_input', '_start', '_end', '_step'
    _add_args_slots = False


class Swapcase(StringOp):
    _args = '_input',
    _add_args_slots = False


class Title(StringOp):
    _args = '_input',
    _add_args_slots = False


class Zfill(StringOp):
    _args = '_input', '_width'
    _add_args_slots = False


class Strptime(StringOp):
    _args = '_input', '_date_format'
    _add_args_slots = False


class Isalnum(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isalpha(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isdigit(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isspace(StringOp):
    _args = '_input',
    _add_args_slots = False


class Islower(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isupper(StringOp):
    _args = '_input',
    _add_args_slots = False


class Istitle(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isnumeric(StringOp):
    _args = '_input',
    _add_args_slots = False


class Isdecimal(StringOp):
    _args = '_input',
    _add_args_slots = False


class StringToDict(StringOp):
    _args = '_input', '_item_delim', '_kv_delim'
    _add_args_slots = False


def _string_op(expr, output_expr_cls, output_type=None, **kwargs):
    output_type = output_type or types.string
    if isinstance(expr, (StringSequenceExpr, StringScalar)):
        is_sequence = isinstance(expr, StringSequenceExpr)

        if is_sequence:
            return output_expr_cls(_data_type=output_type, _input=expr, **kwargs)
        else:
            return output_expr_cls(_value_type=output_type, _input=expr, **kwargs)


def _capitalize(expr):
    """
    Convert strings in the Sequence or string scalar to be capitalized. Equivalent to str.capitalize().

    :param expr:
    :return: sequence or scalar
    """

    return _string_op(expr, Capitalize)


def _cat(expr, others, sep=None, na_rep=None):
    if isinstance(others, six.string_types):
        raise ValueError('Did you mean to supply a `sep` keyword?')
    return _string_op(expr, CatStr, _others=others, _sep=sep, _na_rep=na_rep)


def _contains(expr, pat, case=True, flags=0, regex=True):
    """
    Return boolean sequence whether given pattern/regex is contained in each string in the sequence

    :param expr: sequence or scalar
    :param pat: Character sequence or regular expression
    :param case: If True, case sensitive
    :type case: bool
    :param flags: re module flags, e.g. re.IGNORECASE
    :param regex: If True use regex, otherwise use string finder
    :return: sequence or scalar
    """

    if regex and isinstance(pat, six.string_types):
        import re
        try:
            re.compile(pat, flags=flags)
        except:
            raise ValueError('Failed to compile regular expression, '
                             'please check re.compile("{0}")'.format(pat))

    return _string_op(expr, Contains, output_type=types.boolean,
                      _pat=pat, _case=case, _flags=flags, _regex=regex)


def _count(expr, pat, flags=0):
    """
    Count occurrences of pattern in each string of the sequence or scalar

    :param expr: sequence or scalar
    :param pat: valid regular expression
    :param flags: re module flags, e.g. re.IGNORECASE
    :return:
    """
    return _string_op(expr, Count, output_type=types.int64,
                      _pat=pat, _flags=flags)


def _endswith(expr, pat):
    """
    Return boolean sequence or scalar indicating whether each string in the sequence or scalar
    ends with passed pattern. Equivalent to str.endswith().

    :param expr:
    :param pat: Character sequence
    :return: sequence or scalar
    """

    return _string_op(expr, Endswith, output_type=types.boolean, _pat=pat)


def _startswith(expr, pat):
    """
    Return boolean sequence or scalar indicating whether each string in the sequence or scalar
    starts with passed pattern. Equivalent to str.startswith().

    :param expr:
    :param pat: Character sequence
    :return: sequence or scalar
    """

    return _string_op(expr, Startswith, output_type=types.boolean, _pat=pat)


def _extract(expr, pat, flags=0, group=0):
    """
    Find group in each string in the Series using passed regular expression.

    :param expr:
    :param pat: Pattern or regular expression
    :param flags: re module, e.g. re.IGNORECASE
    :param group: if None as group 0
    :return: sequence or scalar
    """

    return _string_op(expr, Extract, _pat=pat, _flags=flags, _group=group)


def _find(expr, sub, start=0, end=None):
    """
    Return lowest indexes in each strings in the sequence or scalar
    where the substring is fully contained between [start:end]. Return -1 on failure.
    Equivalent to standard str.find().

    :param expr:
    :param sub: substring being searched
    :param start: left edge index
    :param end: right edge index
    :return: sequence or scalar
    """

    return _string_op(expr, Find, output_type=types.int64,
                      _sub=sub, _start=start, _end=end)


def _rfind(expr, sub, start=0, end=None):
    """
    Return highest indexes in each strings in the sequence or scalar
    where the substring is fully contained between [start:end]. Return -1 on failure.
    Equivalent to standard str.rfind().

    :param expr:
    :param sub:
    :param start:
    :param end:
    :return: sequence or scalar
    """

    return _string_op(expr, RFind, output_type=types.int64,
                      _sub=sub, _start=start, _end=end)


def _replace(expr, pat, repl, n=-1, case=True, flags=0, regex=True):
    """
    Replace occurrence of pattern/regex in the sequence or scalar with some other string.
    Equivalent to str.replace()

    :param expr:
    :param pat: Character sequence or regular expression
    :param repl: Replacement
    :param n: Number of replacements to make from start
    :param case: if True, case sensitive
    :param flags: re module flag, e.g. re.IGNORECASE
    :return: sequence or scalar
    """

    return _string_op(expr, Replace, _pat=pat, _repl=repl,
                      _n=n, _case=case, _flags=flags, _regex=regex)


def _get(expr, index):
    """
    Extract element from lists, tuples, or strings in each element in the sequence or scalar

    :param expr:
    :param index: Integer index(location)
    :return: sequence or scalar
    """

    return _string_op(expr, Get, _index=index)


def _join(expr, sep):
    """
    Join lists contained as elements in the Series/Index with passed delimiter.
    Equivalent to str.join().

    :param expr:
    :param sep: Delimiter
    :return: sequence or scalar
    """

    return _string_op(expr, Join, _sep=sep)


def _len(expr):
    """
    Compute length of each string in the sequence or scalar

    :param expr:
    :return: lengths
    """
    return _string_op(expr, Len, output_type=types.int64)


def _ljust(expr, width, fillchar=' '):
    """
    Filling right side of strings in the sequence or scalar with an additional character.
    Equivalent to str.ljust().

    :param expr:
    :param width: Minimum width of resulting string; additional characters will be filled with `fillchar`
    :param fillchar: Additional character for filling, default is whitespace.
    :return: sequence or scalar
    """
    return _string_op(expr, Ljust, _width=width, _fillchar=fillchar)


def _rjust(expr, width, fillchar=' '):
    """
    Filling left side of strings in the sequence or scalar with an additional character.
    Equivalent to str.rjust().

    :param expr:
    :param width: Minimum width of resulting string; additional characters will be filled with `fillchar`
    :param fillchar: Additional character for filling, default is whitespace.
    :return: sequence or scalar
    """

    return _string_op(expr, Rjust, _width=width, _fillchar=fillchar)


def _lower(expr):
    """
    Convert strings in the sequence or scalar lowercase. Equivalent to str.lower().

    :param expr:
    :return: sequence or scalar
    """
    return _string_op(expr, Lower)


def _upper(expr):
    """
    Convert strings in the sequence or scalar uppercase. Equivalent to str.upper().

    :param expr:
    :return: sequence or scalar
    """

    return _string_op(expr, Upper)


def _lstrip(expr, to_strip=None):
    """
    Strip whitespace (including newlines) from each string in the sequence or scalar from left side.
    Equivalent to str.lstrip().

    :param expr:
    :param to_strip:
    :return: sequence or sclaar
    """

    return _string_op(expr, Lstrip, _to_strip=to_strip)


def _rstrip(expr, to_strip=None):
    """
    Strip whitespace (including newlines) from each string in the sequence or scalar from right side.
    Equivalent to str.rstrip().

    :param expr:
    :param to_strip:
    :return: sequence or scalar
    """

    return _string_op(expr, Rstrip, _to_strip=to_strip)


def _split(expr, pat=None, n=-1):
    """
    Split each string (a la re.split) in the Series/Index by given pattern, propagating NA values.
    Equivalent to str.split().

    :param expr:
    :param pat: Separator to split on. If None, splits on whitespace
    :param n: not supported right now
    :return: list sequence or scalar
    """

    return _string_op(expr, Split, output_type=types.List(types.string),
                      _pat=pat, _n=n)


def _rsplit(expr, pat=None, n=-1):
    """
    Split each string in the Series/Index by the given delimiter string,
    starting at the end of the string and working to the front.
    Equivalent to str.rsplit().

    :param expr:
    :param pat: Separator to split on. If None, splits on whitespace
    :param n: None, 0 and -1 will be interpreted as return all splits
    :return: sequence or scalar
    """

    return _string_op(expr, RSplit, output_type=types.List(types.string),
                      _pat=pat, _n=n)


def _strip(expr, to_strip=None):
    """
    Strip whitespace (including newlines) from each string in the sequence or scalar from left and right sides.
    Equivalent to str.strip().

    :param expr:
    :param to_strip:
    :return: sequence or scalar
    """

    return _string_op(expr, Strip, _to_strip=to_strip)


def _pad(expr, width, side='left', fillchar=' '):
    """
    Pad strings in the sequence or scalar with an additional character to specified side.

    :param expr:
    :param width: Minimum width of resulting string; additional characters will be filled with spaces
    :param side: {‘left’, ‘right’, ‘both’}, default ‘left’
    :param fillchar: Additional character for filling, default is whitespace
    :return: sequence or scalar
    """

    if not isinstance(fillchar, six.string_types):
        msg = 'fillchar must be a character, not {0}'
        raise TypeError(msg.format(type(fillchar).__name__))

    if len(fillchar) != 1:
        raise TypeError('fillchar must be a character, not str')

    if side not in ('left', 'right', 'both'):
        raise ValueError('Invalid side')

    return _string_op(expr, Pad, _width=width, _side=side, _fillchar=fillchar)


def _repeat(expr, repeats):
    """
    Duplicate each string in the sequence or scalar by indicated number of times.

    :param expr:
    :param repeats: times
    :return: sequence or scalar
    """

    return _string_op(expr, Repeat, _repeats=repeats)


def _slice(expr, start=None, stop=None, step=None):
    """
    Slice substrings from each element in the sequence or scalar

    :param expr:
    :param start: int or None
    :param stop: int or None
    :param step: int or None
    :return: sliced
    """

    return _string_op(expr, Slice, _start=start, _end=stop, _step=step)


def _getitem(expr, item):
    if isinstance(item, six.integer_types) or \
            (isinstance(item, (SequenceExpr, Scalar)) and isinstance(item.dtype, types.Integer)):
        return _get(expr, item)
    elif isinstance(item, slice):
        return _slice(expr, start=item.start, stop=item.stop, step=item.step)
    else:
        raise TypeError('Unknown argument: %r' % item)


def _swapcase(expr):
    """
    Convert strings in the sequence or scalar to be swapcased. Equivalent to str.swapcase().

    :param expr:
    :return: converted
    """

    return _string_op(expr, Swapcase)


def _title(expr):
    """
    Convert strings in the sequence or scalar to titlecase. Equivalent to str.title().


    :param expr:
    :return: converted
    """

    return _string_op(expr, Title)


def _zfill(expr, width):
    """
    Filling left side of strings in the sequence or scalar with 0. Equivalent to str.zfill().

    :param expr:
    :param width: Minimum width of resulting string; additional characters will be filled with 0
    :return: filled
    """

    return _string_op(expr, Zfill, _width=width)


def _strptime(expr, date_format):
    """
    Return datetimes specified by date_format,
    which supports the same string format as the python standard library.
    Details of the string format can be found in python string format doc

    :param expr:
    :param date_format: date format string (e.g. “%Y-%m-%d”)
    :type date_format: str
    :return:
    """

    return _string_op(expr, Strptime, _date_format=date_format,
                      output_type=types.datetime)


def _isalnum(expr):
    """
    Check whether all characters in each string in the sequence or scalar are alphanumeric.
    Equivalent to str.isalnum().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isalnum, output_type=types.boolean)


def _isalpha(expr):
    """
    Check whether all characters in each string in the sequence or scalar are alphabetic.
    Equivalent to str.isalpha().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isalpha, output_type=types.boolean)


def _isdigit(expr):
    """
    Check whether all characters in each string in the sequence or scalar are digits.
    Equivalent to str.isdigit().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isdigit, output_type=types.boolean)


def _isspace(expr):
    """
    Check whether all characters in each string in the sequence or scalar are whitespace.
    Equivalent to str.isspace().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isspace, output_type=types.boolean)


def _islower(expr):
    """
    Check whether all characters in each string in the sequence or scalar are lowercase.
    Equivalent to str.islower().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Islower, output_type=types.boolean)


def _isupper(expr):
    """
    Check whether all characters in each string in the sequence or scalar are uppercase.
    Equivalent to str.isupper().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isupper, output_type=types.boolean)


def _istitle(expr):
    """
    Check whether all characters in each string in the sequence or scalar are titlecase.
    Equivalent to str.istitle().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Istitle, output_type=types.boolean)


def _isnumeric(expr):
    """
    Check whether all characters in each string in the sequence or scalar are numeric.
    Equivalent to str.isnumeric().


    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isnumeric, output_type=types.boolean)


def _isdecimal(expr):
    """
    Check whether all characters in each string in the sequence or scalar are decimal.
    Equivalent to str.isdecimal().

    :param expr:
    :return: boolean sequence or scalar
    """

    return _string_op(expr, Isdecimal, output_type=types.boolean)


def _todict(expr, item_delim=',', kv_delim='='):
    """
    Convert the string sequence / expr into a string dict given item and key-value delimiters.

    :param expr:
    :param item_delim: delimiter between data items
    :param kv_delim: delimiter between keys and values
    :return: dict sequence or scalar
    """
    return _string_op(expr, StringToDict, _item_delim=item_delim, _kv_delim=kv_delim,
                      output_type=types.Dict(types.string, types.string))


_string_methods = dict(
    capitalize=_capitalize,
    contains=_contains,
    count=_count,
    endswith=_endswith,
    startswith=_startswith,
    extract=_extract,
    find=_find,
    rfind=_rfind,
    replace=_replace,
    get=_get,
    len=_len,
    ljust=_ljust,
    rjust=_rjust,
    lower=_lower,
    upper=_upper,
    lstrip=_lstrip,
    rstrip=_rstrip,
    split=_split,
    strip=_strip,
    pad=_pad,
    repeat=_repeat,
    slice=_slice,
    __getitem__=_getitem,
    swapcase=_swapcase,
    title=_title,
    zfill=_zfill,
    strptime=_strptime,
    isalnum=_isalnum,
    isalpha=_isalpha,
    isdigit=_isdigit,
    isspace=_isspace,
    islower=_islower,
    isupper=_isupper,
    istitle=_istitle,
    isnumeric=_isnumeric,
    isdecimal=_isdecimal,
    todict=_todict,
)


utils.add_method(StringSequenceExpr, _string_methods)
utils.add_method(StringScalar, _string_methods)
utils.add_method(StringScalar, {'cat': _cat})



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from ...compat import six, Iterable
from ...serializers import JSONSerializableModel, JSONNodeField
from ..utils import ML_ARG_PREFIX

"""
Exporter
"""


def get_append_id_selected_cols(expr, param_value):
    if param_value:
        if isinstance(param_value, (list, tuple, set)):
            return ','.join(param_value)
        return param_value
    else:
        return ','.join(f.name for f in getattr(expr, ML_ARG_PREFIX + 'input')._ml_fields)


def get_modify_abnormal_json_string(expr, param_value):
    if isinstance(param_value, JSONSerializableModel):
        return json.dumps([param_value.serial(), ], separators=(',', ':'))
    if isinstance(param_value, Iterable) and not isinstance(param_value, six.string_types):
        return json.dumps([obj.serial() if isinstance(obj, JSONSerializableModel) else obj for obj in param_value],
                          separators=(',', ':'))
    else:
        return param_value


"""
Output Schemas
"""


def binning_predict_output(params, fields):
    out_cols = params['metaColNames'].split(',') if 'metaColNames' in params else []
    input_ml_fields = fields['feature']
    if not out_cols:
        out_cols = list(six.iterkeys(fields['feature']))
    return ','.join('%s: %s' % (col, input_ml_fields[col]) for col in out_cols) + ', prediction_score: double, ' +\
           'prediction_prob: double, prediction_detail: string'


"""
Modify Abnormal JSON classes
"""


class ReplaceNull(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    replace = JSONNodeField('replace')

    def __init__(self, col, replace, **kwargs):
        kwargs.update(dict(col=col, type='null', replace=replace))
        super(ReplaceNull, self).__init__(**kwargs)


class ReplaceEmpty(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    replace = JSONNodeField('replace')

    def __init__(self, col, replace, **kwargs):
        kwargs.update(dict(col=col, type='empty', replace=replace))
        super(ReplaceEmpty, self).__init__(**kwargs)


class ReplaceNullEmpty(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    replace = JSONNodeField('replace')

    def __init__(self, col, replace, **kwargs):
        kwargs.update(dict(col=col, type='null-empty', replace=replace))
        super(ReplaceNullEmpty, self).__init__(**kwargs)


class ReplaceCustom(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    original = JSONNodeField('original')
    replace = JSONNodeField('replace')

    def __init__(self, col, original, replace, **kwargs):
        kwargs.update(dict(col=col, type='user-defined', original=original, replace=replace))
        super(ReplaceCustom, self).__init__(**kwargs)


class ReplacePercentile(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    low_range = JSONNodeField('low_range')
    high_range = JSONNodeField('high_range')
    low_replace = JSONNodeField('low_replace')
    high_replace = JSONNodeField('high_replace')

    def __init__(self, col, low_range=None, low_replace=None, high_range=None, high_replace=None, **kwargs):
        kwargs.update(dict(col=col, type='percentile', low_range=low_range, low_replace=low_replace,
                           high_range=high_range, high_replace=high_replace))
        super(ReplacePercentile, self).__init__(**kwargs)


class ReplaceConfidence(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    confidence = JSONNodeField('confidence')
    low_replace = JSONNodeField('low_replace')
    high_replace = JSONNodeField('high_replace')

    def __init__(self, col, confidence, low_replace=None, high_replace=None, **kwargs):
        kwargs.update(dict(col=col, type='zscore', confidence=confidence, low_replace=low_replace,
                           high_replace=high_replace))
        super(ReplaceConfidence, self).__init__(**kwargs)


class ReplaceZScore(JSONSerializableModel):
    col = JSONNodeField('col')
    type = JSONNodeField('type')
    low_range = JSONNodeField('low_range')
    high_range = JSONNodeField('high_range')
    low_replace = JSONNodeField('low_replace')
    high_replace = JSONNodeField('high_replace')

    def __init__(self, col, low_range=None, low_replace=None, high_range=None, high_replace=None, **kwargs):
        kwargs.update(dict(col=col, type='zscore', low_range=low_range, low_replace=low_replace,
                           high_range=high_range, high_replace=high_replace))
        super(ReplaceZScore, self).__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import uuid
import warnings
from collections import OrderedDict
from itertools import chain

import numpy as np
import oss2
from mars.tensor.utils import decide_chunk_sizes

from ...df import CollectionExpr
from ...compat import six
from ...df.backends.engine import get_default_engine
from ...df.backends.odpssql.types import df_type_to_odps_type
from ...df.backends.pd.types import df_type_to_np_type
from ...models import TableSchema, Table
from ...df import Scalar
from ...utils import to_str
from ...errors import ODPSError
from ...types import PartitionSpec
from ... import types as odps_types
from .tensor import read_coo
from .io import open as fs_open


np_int_types = map(np.dtype, [np.int_, np.int8, np.int16, np.int32, np.int64])
np_float_types = map(np.dtype, [np.float_, np.float16, np.float32, np.float64])
np_to_odps_types = dict(
    [(t, odps_types.bigint) for t in np_int_types]
    + [(t, odps_types.double) for t in np_float_types]
)


logger = logging.getLogger(__name__)


def _clean_oss_object(
    path, access_id=None, secret_access_key=None, endpoint=None, bucket_name=None
):
    bucket = oss2.Bucket(oss2.Auth(access_id, secret_access_key), endpoint, bucket_name)
    while True:
        keys = [l.key for l in oss2.ObjectIterator(bucket, prefix=path + "/")]
        if not keys:
            return
        bucket.batch_delete_objects(keys)


def _copy_to_ext_table(
    expr,
    odps,
    dim_cols,
    val_col,
    chunks,
    partitions=None,
    oss_path=None,
    handler=None,
    hints=None,
    **oss_opts
):
    column_names = list(dim_cols) + [val_col]
    column_types = [df_type_to_odps_type(expr.schema[n].type) for n in column_names]

    if partitions is not None:
        partitions = (
            [partitions]
            if not isinstance(partitions, (list, tuple, set))
            else partitions
        )
        partition_types = ["string"] * len(partitions)
        target_schema = TableSchema.from_lists(
            column_names, column_types, partitions, partition_types
        )
        column_names.extend(list(partitions))
    else:
        target_schema = TableSchema.from_lists(column_names, column_types)

    _clean_oss_object(oss_path + "/data", **oss_opts)

    if not isinstance(chunks, (list, tuple)):
        chunks = [chunks] * len(dim_cols)
    ext_table_name = "mars_ext_%s" % str(uuid.uuid4()).replace("-", "_")
    ext_table = odps.create_table(
        ext_table_name,
        target_schema,
        storage_handler="com.aliyun.odps.mars.ChunkStorageHandler",
        serde_properties={
            "mars.chunk_sizes": ",".join("%s" % ch for ch in chunks),
            "mars.parquet.page_size": "%s" % (128 * 1024),
            "mars.parquet.row_group_size": "%s" % (128 * 1024),
        },
        location="oss://%s:%s@%s/%s/%s/data"
        % (
            oss_opts.get("access_id"),
            oss_opts.get("secret_access_key"),
            oss_opts.get("endpoint").split("://")[1],
            oss_opts.get("bucket_name"),
            oss_path,
        ),
        resources=handler,
        lifecycle=14,
    )
    if "/" in handler:
        assert handler in ext_table.resources
    if partitions:
        shuffled = expr.select(column_names).reshuffle(
            partitions, sort=partitions + dim_cols
        )
    else:
        shuffled = expr.select(column_names).reshuffle(dim_cols, sort=dim_cols)
    logger.debug("writing to external table %s" % ext_table_name)

    odps_hints = {
        "odps.sql.reshuffle.dynamicpt": False,
        "odps.isolation.session.enable": True,
    }
    if hints:
        odps_hints.update(hints)
    shuffled.persist(
        ext_table.name, overwrite=False, partitions=partitions, hints=odps_hints
    )


def _write_shape_to_oss(shape, path, **oss_opts):
    shapes_string = ",".join("%s" % s for s in shape)
    with fs_open(path, "wb", **oss_opts) as out_file:
        out_file.write(shapes_string)


def _read_shape_from_oss(path, **oss_opts):
    with fs_open(path, "rb", **oss_opts) as file:
        shape_string = to_str(file.read())
    return tuple(int(i) for i in shape_string.split(","))


def _as_scalar(s):
    return s.item() if hasattr(s, "item") else s


def _get_shapes(expr, dim_columns, partitions=None, hints=None):
    if partitions is None:
        shapes = expr[
            list((expr[col].max() + 1).rename(col) for col in dim_columns)
        ].execute(hints=hints)
        return tuple(_as_scalar(s) for s in shapes[0])
    else:
        odps_hints = {"odps.task.merge.enabled": False}
        if hints:
            odps_hints.update(odps_hints)
        shapes = (
            expr.groupby(partitions)
            .agg(list((expr[col].max() + 1).rename(col) for col in dim_columns))
            .execute(hints=odps_hints)
        )
        return dict(
            (
                ",".join(tuple(record[partitions])),
                tuple(_as_scalar(s) for s in record[dim_columns]),
            )
            for record in shapes
        )


def _preprocess(expr, table_name, columns, partitions=None, hints=None):
    if partitions:
        expr_grouped = expr.groupby(partitions)
        expr_ranked = expr[
            [
                (expr_grouped[col].dense_rank() - 1).rename(col + "_continuous")
                for col in columns
            ]
        ]
    else:
        expr_ranked = expr[
            [
                (expr.groupby(Scalar(1))[col].dense_rank() - 1).rename(
                    col + "_continuous"
                )
                for col in columns
            ]
        ]
    logger.debug("creating new table %s")
    expr[expr, expr_ranked].persist(table_name, partitions=partitions, hints=hints)


def _get_df_partitions(df, partitions):
    if (
        df._source_data is not None
        and isinstance(df._source_data, Table)
        and set([p.name for p in df._source_data.table_schema.partitions]) == set(partitions)
    ):
        return list(p.partition_spec for p in df._source_data.partitions)
    else:
        return df[partitions].distinct().execute()


def _decide_chunk_sizes(shape, chunk_size, dtype):
    if not isinstance(shape, dict):
        nsplits = decide_chunk_sizes(shape, chunk_size, dtype.itemsize)
        return [int(ns[0]) for ns in nsplits]
    else:
        result_chunks = [None] * len(shape)
        for s in shape.values():
            nsplits = decide_chunk_sizes(s, chunk_size, dtype.itemsize)
            ns = [int(n[0]) for n in nsplits]
            for i, n in enumerate(ns):
                if result_chunks[i] is None or result_chunks[i] > n:
                    result_chunks[i] = n
        return result_chunks


def to_mars_tensor_via_oss(
    expr,
    dim_columns,
    value_column,
    chunks=None,
    chunk_size=None,
    oss_access_id=None,
    oss_access_key=None,
    oss_endpoint=None,
    oss_bucket_name=None,
    oss_path=None,
    odps=None,
    discontinuous_columns=None,
    new_table_name=None,
    partitions=None,
    shape=None,
    sparse=False,
    oss_file_exist=False,
    handler="public/resources/chunkwriter.jar",
    hints=None,
):
    if (discontinuous_columns is not None) != (new_table_name is not None):
        raise ValueError(
            "if some dimension columns are discontinuous,"
            "preprocessed table name should be provided"
        )
    if oss_path is None:
        raise ValueError("oss_path should be provided")

    oss_prefix = "oss://%s/" % oss_bucket_name
    if oss_path.startswith(oss_prefix):
        oss_path = oss_path[len(oss_prefix) :]

    if sparse and (shape is None):
        raise ValueError("shape is necessary if target tensor is sparse")
    if odps is None:
        odps = get_default_engine(expr)._odps
        if odps is None:
            raise ODPSError("ODPS entrance should be provided")
    if isinstance(partitions, six.string_types):
        partitions = [partitions]
    if discontinuous_columns is not None:
        if not odps.exist_table(new_table_name):
            _preprocess(
                expr, new_table_name, discontinuous_columns, partitions, hints=hints
            )
        expr = odps.get_table(new_table_name).to_df()
        disordered_cols = discontinuous_columns or []
        dim_columns = [
            col + "_continuous" if col in disordered_cols else col
            for col in dim_columns
        ]
    oss_opts = dict(
        endpoint=oss_endpoint,
        bucket_name=oss_bucket_name,
        access_id=oss_access_id,
        secret_access_key=oss_access_key,
    )
    dtype = df_type_to_np_type(expr[value_column].dtype)

    if shape is not None and isinstance(shape, dict):
        new_shape = dict()
        for k, v in shape.items():
            p_spec = PartitionSpec(k)
            p_key = ",".join([p_spec[p] for p in partitions])
            new_shape[p_key] = v
        shape = new_shape

    # get shape
    partition_values = None
    if not oss_file_exist:
        shape = shape or _get_shapes(expr, dim_columns, partitions, hints=hints)
    if partitions is None:
        shape_path = "oss://%s/meta/shape" % oss_path
        if shape is None:
            shape = _read_shape_from_oss(shape_path, **oss_opts)
        if not oss_file_exist:
            _write_shape_to_oss(shape, shape_path, **oss_opts)
    else:
        partition_values = _get_df_partitions(expr, partitions)
        new_shape = dict()
        for partition in partition_values:
            partition_path = "/".join("%s=%s" % (p, partition[p]) for p in partitions)
            shape_path = "oss://%s/meta/%s/shape" % (oss_path, partition_path)
            partition_key = ",".join(partition[p] for p in partitions)
            if shape is None:
                partition_shape = _read_shape_from_oss(shape_path, **oss_opts)
                new_shape[partition_key] = partition_shape
            else:
                partition_shape = shape[partition_key]
            if not oss_file_exist:
                _write_shape_to_oss(partition_shape, shape_path, **oss_opts)
        if shape is None:
            shape = new_shape

    if chunks is not None:
        warnings.warn(
            "chunks is deprecated, use chunk_size instead", DeprecationWarning
        )

    chunk_size = chunk_size if chunk_size is not None else chunks
    if chunk_size is None:
        chunk_size = _decide_chunk_sizes(shape, chunk_size, dtype)

    if not oss_file_exist:
        _copy_to_ext_table(
            expr,
            odps,
            dim_columns,
            value_column,
            chunk_size,
            oss_path=oss_path,
            partitions=partitions,
            handler=handler,
            hints=hints,
            **oss_opts
        )

    if partitions is None:
        # for non-partitioned table
        path = "oss://%s/data/*.parquet" % oss_path
        tensor = read_coo(
            path,
            dim_columns,
            value_column,
            shape=shape,
            chunk_size=chunk_size,
            dtype=dtype,
            sparse=sparse,
            **oss_opts
        )
        return tensor
    else:
        # for partitioned table
        tensors = OrderedDict()
        for partition in partition_values:
            partition_path = "/".join("%s=%s" % (p, partition[p]) for p in partitions)
            path = "oss://%s/data/%s/*.parquet" % (oss_path, partition_path)
            partition_key = ",".join(partition[p] for p in partitions)
            tensor = read_coo(
                path,
                dim_columns,
                value_column,
                shape=shape[partition_key],
                chunk_size=chunk_size,
                dtype=dtype,
                sparse=sparse,
                **oss_opts
            )
            partition_spec = ",".join("%s=%s" % (p, partition[p]) for p in partitions)
            tensors[partition_spec] = tensor
        return tensors


def persist_tensor_via_oss(odps, *args, **kwargs):
    from mars.session import Session
    from .tensor.datastore import write_coo

    session = kwargs.pop("session", Session.default_or_local())
    oss_endpoint = kwargs.pop("oss_endpoint")
    oss_access_id = kwargs.pop("oss_access_id")
    oss_access_key = kwargs.pop("oss_access_key")
    oss_bucket_name = kwargs.pop("oss_bucket_name")
    oss_path = kwargs.pop("oss_path")

    oss_prefix = "oss://%s/" % oss_bucket_name
    if oss_path.startswith(oss_prefix):
        oss_path = oss_path[len(oss_prefix) :]

    oss_opts = dict(
        endpoint=oss_endpoint,
        bucket_name=oss_bucket_name,
        access_id=oss_access_id,
        secret_access_key=oss_access_key,
    )

    tensor, table_name, dim_columns, value_column = args
    oss_dir = "oss://%s" % oss_path
    _clean_oss_object(oss_path, **oss_opts)

    t_type = None
    partitions = None

    # submit tensor to mars cluster
    tensors = []
    if isinstance(tensor, dict):
        for p, t in tensor.items():
            if t_type is None:
                t_type = t.dtype
            p_spec = PartitionSpec(p)
            if partitions is None:
                partitions = p_spec.keys
            else:
                if set(partitions) != set(p_spec.keys):
                    raise TypeError("all tensors partitions name must be the same.")

            if t.ndim > len(dim_columns):
                raise TypeError("tensor dimensions cannot more than dim_columns length")

            # write shape to oss
            shape_path = "%s/meta/%s/shape" % (oss_dir, p.replace(",", "/"))
            _write_shape_to_oss(t.shape, shape_path, **oss_opts)

            # write data to oss
            data_path = "%s/data/%s" % (oss_dir, p.replace(",", "/"))
            writer_tensor = write_coo(
                t, data_path, dim_columns, value_column, global_index=True, **oss_opts
            )
            tensors.append(writer_tensor)

        session.run(tensors)
    else:
        shape_path = oss_dir + "/meta/shape"
        _write_shape_to_oss(tensor.shape, shape_path, **oss_opts)

        t_type = tensor.dtype
        data_path = oss_dir + "/data"
        writer_tensor = write_coo(
            tensor, data_path, dim_columns, value_column, global_index=True, **oss_opts
        )
        session.run(writer_tensor)

    # persist to odps table
    ext_table_name = "mars_persist_ext_%s" % str(uuid.uuid4()).replace("-", "_")
    column_types = ["bigint"] * len(dim_columns) + [np_to_odps_types[t_type]]
    ext_column_types = ["bigint"] * (2 * len(dim_columns)) + [np_to_odps_types[t_type]]
    column_names = dim_columns + [value_column]
    ext_column_names = list(chain(*([c, "global_" + c] for c in dim_columns))) + [
        value_column
    ]
    if partitions:
        if isinstance(partitions, six.string_types):
            partitions = [partitions]
        target_schema = TableSchema.from_lists(
            column_names, column_types, partitions, ["string"] * len(partitions)
        )
        ext_schema = TableSchema.from_lists(
            ext_column_names, ext_column_types, partitions, ["string"] * len(partitions)
        )
    else:
        target_schema = TableSchema.from_lists(column_names, column_types)
        ext_schema = TableSchema.from_lists(ext_column_names, ext_column_types)

    ext_table = odps.create_table(
        ext_table_name,
        ext_schema,
        external_stored_as="PARQUET",
        location="oss://%s:%s@%s/%s/%s/data"
        % (
            oss_opts["access_id"],
            oss_opts["secret_access_key"],
            oss_opts["endpoint"].split("://")[1],
            oss_opts["bucket_name"],
            oss_path,
        ),
    )
    if partitions:
        for partition in tensor.keys():
            ext_table.create_partition(partition)
    odps.create_table(table_name, target_schema, if_not_exists=True)
    ext_df = ext_table.to_df()
    fields = [
        ext_df["global_" + f].rename(f) for f in target_schema.names[:-1]
    ] + target_schema.names[-1:]
    if partitions:
        fields = fields + partitions
        ext_df[fields].persist(table_name, partitions=partitions)
    else:
        ext_df[fields].persist(table_name)


CollectionExpr.to_mars_tensor_via_oss = to_mars_tensor_via_oss



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""

import struct

from . import input_stream
from . import wire_format


class Decoder(object):
    """Decodes logical protocol buffer fields from the wire."""

    def __init__(self, input):
        """Initializes the decoder to read from input stream.
        """
        self._stream = input_stream.InputStream(input)

    def __len__(self):
        """Returns the 0-indexed position in |s|."""
        return self._stream.position()

    def read_field_number_and_wire_type(self):
        """Reads a tag from the wire. Returns a (field_number, wire_type) pair."""
        tag_and_type = self.read_uint32()
        return wire_format.unpack_tag(tag_and_type)

    def read_uint32(self):
        """Reads and returns an signed, varint-encoded, 32-bit integer."""
        return self._stream.read_var_uint32()

    def read_uint64(self):
        """Reads and returns an signed, varint-encoded,64-bit integer."""
        return self._stream.read_var_uint64()

    def read_sint32(self):
        """Reads and returns a signed, zigzag-encoded, varint-encoded,
        32-bit integer."""
        return wire_format.zig_zag_decode(self._stream.read_var_uint32())

    def read_sint64(self):
        """Reads and returns a signed, zigzag-encoded, varint-encoded,
        64-bit integer."""
        return wire_format.zig_zag_decode(self._stream.read_var_uint64())

    def read_fixed32(self):
        """Reads and returns an unsigned, fixed-width, 32-bit integer."""
        return self._stream.read_little_endian32()

    def read_fixed64(self):
        """Reads and returns an unsigned, fixed-width, 64-bit integer."""
        return self._stream.read_little_endian64()

    def read_sfixed32(self):
        """Reads and returns a signed, fixed-width, 32-bit integer."""
        value = self._stream.read_little_endian32()
        if value >= (1 << 31):
            value -= (1 << 32)
        return value

    def read_sfixed64(self):
        """Reads and returns a signed, fixed-width, 64-bit integer."""
        value = self._stream.read_little_endian64()
        if value >= (1 << 63):
            value -= (1 << 64)
        return value

    def read_float(self):
        """Reads and returns a 4-byte floating-point number."""
        serialized = self._stream.read_string(4)
        return struct.unpack('f', serialized)[0]

    def read_double(self):
        """Reads and returns an 8-byte floating-point number."""
        serialized = self._stream.read_string(8)
        return struct.unpack('d', serialized)[0]

    def read_bool(self):
        """Reads and returns a bool."""
        i = self._stream.read_var_uint32()
        return bool(i)

    def read_enum(self):
        """Reads and returns an enum value."""
        return self._stream.read_var_uint32()

    def read_string(self):
        """Reads and returns a length-delimited string."""
        length = self._stream.read_var_uint32()
        return self._stream.read_string(length)

    def ReadBytes(self):
        """Reads and returns a length-delimited byte sequence."""
        return self.read_string()



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *


cdef class Encoder:
    cdef int _pos
    cdef int _buf_len
    cdef int _last_error
    cdef char *_buf_ptr

    cdef init(self, char *buf_ptr, int buf_len)

    cdef int position(self) nogil
    cdef int get_last_error(self) nogil
    cdef void set_last_error(self, int errno) nogil

    cdef int append_tag(self, int field_num, int wire_type) nogil
    cdef int append_sint32(self, int32_t value) nogil
    cdef int append_uint32(self, uint32_t value) nogil
    cdef int append_sint64(self, int64_t value) nogil
    cdef int append_uint64(self, uint64_t value) nogil
    cdef int append_bool(self, bint value) nogil
    cdef int append_float(self, float value) nogil
    cdef int append_double(self, double value) nogil
    cdef int append_string(self, const char *ptr, int value_len) nogil

    cdef int _set_varint32(self, int32_t varint) nogil
    cdef int _set_varint64(self, int64_t varint) nogil
    cdef int _set_signed_varint32(self, int32_t varint) nogil
    cdef int _set_signed_varint64(self, int64_t varint) nogil



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pytest

from ...config import options
from ...tests.core import tn

if sys.version_info[0] >= 3:
    from .storage_api import StorageApiArrowClient
else:
    StorageApiArrowClient = None


_test_table_id = 0


@pytest.fixture
def storage_api_client(odps):
    global _test_table_id

    options.always_enable_schema = True

    test_table_name = tn("test_halo_common_table_" + str(_test_table_id))
    _test_table_id += 1
    odps.delete_table(test_table_name, if_exists=True)
    table = odps.create_table(
        test_table_name,
        ("a BIGINT, b BIGINT, c BIGINT, d BIGINT", "pt string"),
        if_not_exists=True,
    )
    try:
        yield StorageApiArrowClient(odps, table)
    finally:
        table.drop(async_=True)
        options.always_enable_schema = False




# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..compat import six
from ..config import options
from ..lib.tblib import pickling_support
from ..readers import AbstractRecordReader
from ..utils import call_with_retry

pickling_support.install()


class TunnelReaderMixin(object):
    def _to_pandas_with_processes(self, start=None, count=None, columns=None, n_process=1):
        import pandas as pd
        import multiprocessing
        from multiprocessing import Pipe

        session_id = self._download_session.id
        start = start or 0
        count = count or self._download_session.count
        count = min(count, self._download_session.count - start)
        try:
            _mp_context = multiprocessing.get_context('fork')
        except ValueError:
            _mp_context = multiprocessing.get_context('spawn')
        except AttributeError:
            # for py27 compatibility
            _mp_context = multiprocessing

        n_process = min(count, n_process)
        split_count = count // n_process + (count % n_process != 0)
        conns = []
        for i in range(n_process):
            parent_conn, child_conn = Pipe()

            p = _mp_context.Process(
                target=self._get_process_split_reader(columns=columns),
                args=(child_conn, session_id, start, split_count, i),
            )
            p.start()
            start += split_count
            conns.append(parent_conn)

        results = [c.recv() for c in conns]
        splits = sorted(results, key=lambda x: x[0])
        if any(not d[2] for d in splits):
            exc_info = next(d[1] for d in splits if not d[2])
            six.reraise(*exc_info)
        return pd.concat([d[1] for d in splits]).reset_index(drop=True)

    def _get_process_split_reader(self, columns=None):
        raise NotImplementedError

    def _open_and_iter_reader(
        self, start, record_count, step=None, compress=False, columns=None, counter=None
    ):
        raise NotImplementedError

    def _retry_iter_reader(
        self, start, record_count, step=None, compress=False, columns=None
    ):
        end = start + record_count
        retry_num = 0
        while start < end:
            is_yield_err = False
            counter = [0]
            try:
                for rec in self._open_and_iter_reader(
                    start, record_count, step, compress=compress, columns=columns, counter=counter
                ):
                    try:
                        # next record successfully obtained, reset retry counter
                        retry_num = 0
                        yield rec
                    except BaseException:
                        # avoid catching errors caused in yield block
                        is_yield_err = True
                        raise
                break
            except:
                retry_num += 1
                if is_yield_err or retry_num > options.retry_times:
                    raise
            finally:
                start += counter[0]
                record_count -= counter[0]


class TunnelRecordReader(TunnelReaderMixin, AbstractRecordReader):
    def __init__(self, parent, download_session, columns=None):
        self._it = iter(self)
        self._parent = parent
        self._download_session = download_session
        self._column_names = columns

    @property
    def download_id(self):
        return self._download_session.id

    @property
    def count(self):
        return self._download_session.count

    @property
    def status(self):
        return self._download_session.status

    def __iter__(self):
        for record in self.read():
            yield record

    def __next__(self):
        return next(self._it)

    next = __next__

    def _iter(self, start=None, end=None, step=None, compress=False, columns=None):
        count = self._calc_count(start, end, step)
        return self.read(
            start=start, count=count, step=step, compress=compress, columns=columns
        )

    def _open_and_iter_reader(
        self, start, record_count, step=None, compress=False, columns=None, counter=None
    ):
        counter = counter or [0]
        with call_with_retry(
            self._download_session.open_record_reader,
            start, record_count, compress=compress, columns=columns
        ) as reader:
            for record in reader[::step]:
                counter[0] += step
                yield record

    def read(self, start=None, count=None, step=None,
             compress=False, columns=None):
        start = start or 0
        step = step or 1
        max_rec_count = self.count - start
        rec_count = min(max_rec_count, count * step) if count is not None else max_rec_count
        columns = columns or self._column_names

        if rec_count == 0:
            return

        for record in self._retry_iter_reader(
            start, rec_count, step=step, compress=compress, columns=columns
        ):
            yield record

    def to_pandas(self, start=None, count=None, columns=None, n_process=1):
        columns = columns or self._column_names
        if n_process == 1 or self._download_session.count == 0:
            return super(TunnelRecordReader, self).to_pandas(
                start=start, count=count, columns=columns
            )
        else:
            return self._to_pandas_with_processes(
                start=start, count=count, columns=columns, n_process=n_process
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TunnelArrowReader(TunnelReaderMixin):
    def __init__(self, parent, download_session, columns=None):
        self._it = iter(self)
        self._parent = parent
        self._download_session = download_session
        self._column_names = columns

    @property
    def download_id(self):
        return self._download_session.id

    @property
    def count(self):
        return self._download_session.count

    @property
    def status(self):
        return self._download_session.status

    def __iter__(self):
        for batch in self.read():
            yield batch

    def __next__(self):
        return next(self._it)

    next = __next__

    def _open_and_iter_reader(
        self, start, record_count, step=None, compress=False, columns=None, counter=None
    ):
        counter = counter or [0]
        with call_with_retry(
            self._download_session.open_arrow_reader,
            start, record_count, compress=compress, columns=columns
        ) as reader:
            while True:
                batch = reader.read_next_batch()
                if batch is not None:
                    counter[0] += batch.num_rows
                    yield batch
                else:
                    break

    def read(self, start=None, count=None, compress=False, columns=None):
        start = start or 0
        max_rec_count = self.count - start
        rec_count = min(max_rec_count, count) if count is not None else max_rec_count
        columns = columns or self._column_names

        if rec_count == 0:
            return

        for batch in self._retry_iter_reader(
            start, rec_count, compress=compress, columns=columns
        ):
            yield batch

    def read_all(self, start=None, count=None, columns=None):
        start = start or 0
        count = count if count is not None else self.count - start
        columns = columns or self._column_names

        if count == 0:
            from ..tunnel.io.types import odps_schema_to_arrow_schema

            arrow_schema = odps_schema_to_arrow_schema(self.schema)
            return arrow_schema.empty_table()

        with self._download_session.open_arrow_reader(
            start, count, columns=columns
        ) as reader:
            return reader.read()

    def to_pandas(self, start=None, count=None, columns=None, n_process=1):
        start = start or 0
        count = count if count is not None else self.count - start
        columns = columns or self._column_names

        if n_process == 1:
            if count == 0:
                from ..tunnel.io.types import odps_schema_to_arrow_schema

                arrow_schema = odps_schema_to_arrow_schema(self.schema)
                return arrow_schema.empty_table().to_pandas()

            with self._download_session.open_arrow_reader(
                start, count, columns=columns
            ) as reader:
                return reader.to_pandas()
        else:
            return self._to_pandas_with_processes(
                start=start, count=count, columns=columns, n_process=n_process
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import pytest

from odps.tests.core import get_config, drop_test_tables


@pytest.fixture(scope="session")
def odps():
    cfg = get_config()
    return cfg.odps


@pytest.fixture(scope="session")
def odps_daily():
    try:
        return get_config().odps_daily
    except AttributeError:
        pytest.skip("ODPS project with schema not defined")


@pytest.fixture(scope="session")
def odps_with_storage_tier():
    try:
        return get_config().odps_with_storage_tier
    except AttributeError:
        pytest.skip("ODPS project with schema not defined")


@pytest.fixture(scope="session")
def odps_with_schema():
    try:
        return get_config().odps_with_schema
    except AttributeError:
        pytest.skip("ODPS project with schema not defined")


@pytest.fixture(scope="session")
def odps_with_schema_tenant():
    try:
        return get_config().odps_with_schema_tenant
    except AttributeError:
        pytest.skip("ODPS project with schema configured on tenants not defined")


@pytest.fixture(scope="session")
def odps_with_tunnel_quota():
    try:
        return get_config().odps_with_tunnel_quota
    except AttributeError:
        pytest.skip("ODPS project with quota not defined")


@pytest.fixture(scope="session")
def config():
    return get_config()


@pytest.fixture(scope="session")
def delete_test_tables(odps):
    try:
        yield
    finally:
        drop_test_tables(odps)


@pytest.fixture(scope="session")
def tunnel():
    return get_config().tunnel


@pytest.mark.optionalhook
def pytest_html_results_table_html(report, data):
    """
    Hide captured logs for successful tests
    """
    try:
        from py.xml import html
    except ImportError:
        return

    if report.passed:
        del data[:]
        data.append(
            html.div("Logs disabled for passed tests.", **{"class": "log"})
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import serializers, errors, utils
from ..compat import six
from .core import Iterable
from .table import Table


class Tables(Iterable):

    marker = serializers.XMLNodeField('Marker')
    max_items = serializers.XMLNodeField('MaxItems', parse_callback=int)
    tables = serializers.XMLNodesReferencesField(Table, 'Table')

    def _get(self, item):
        return Table(client=self._client, parent=self, name=item)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            table = self._get(item)
        elif isinstance(item, Table):
            table = item
        else:
            return False

        try:
            table.reload()
            return True
        except errors.NoSuchObject:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self, name=None, owner=None, type=None, extended=False):
        """
        :param name: the prefix of table name
        :param owner: owner of the table
        :param type: type of the table
        :param extended: load extended information for table
        :return:
        """
        actions = []
        params = {'expectmarker': 'true'}
        if name is not None:
            params['name'] = name
        if owner is not None:
            params['owner'] = owner
        if type is not None:
            table_type = type.upper() if isinstance(type, str) else type
            table_type = Table.Type(table_type)
            params['type'] = table_type.value
        if extended:
            actions.append("extended")

        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, actions=actions, params=params)

            t = Tables.parse(self._client, resp, obj=self)
            params['marker'] = t.marker

            return t.tables

        while True:
            tables = _it()
            if tables is None:
                break
            for table in tables:
                yield table

    @utils.with_wait_argument
    def create(self, table_name, table_schema, comment=None, if_not_exists=False,
               lifecycle=None, shard_num=None, hub_lifecycle=None, hints=None,
               transactional=False, storage_tier=None, async_=False, **kw):
        project_name = self._parent.project.name
        schema_name = self._get_schema_name()
        sql = Table.gen_create_table_sql(
            table_name, table_schema, comment=comment, if_not_exists=if_not_exists,
            lifecycle=lifecycle, shard_num=shard_num, hub_lifecycle=hub_lifecycle,
            transactional=transactional, project=project_name,
            schema=schema_name, **kw)

        from .tasks import SQLTask

        task = SQLTask(name='SQLCreateTableTask', query=sql)
        hints = hints or {}
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        else:
            hints["odps.namespace.schema"] = "false"
        if storage_tier:
            hints['odps.tiered.storage.enable'] = 'true'
        task.update_sql_settings(hints)
        instance = self._parent.project.instances.create(task=task)

        if not async_:
            instance.wait_for_success()

            return self[table_name]
        else:
            return instance

    def _gen_delete_table_sql(self, table_name, if_exists=False):
        project_name = self._parent.project.name
        schema_name = self._get_schema_name()

        buf = six.StringIO()
        tb = self._get(table_name)
        if tb._getattr("type") is None:
            # if table not loaded, use 'TABLE' type to reduce request
            type_str = 'TABLE'
        elif tb.type == Table.Type.VIRTUAL_VIEW:
            type_str = 'VIEW'
        elif tb.type == Table.Type.MATERIALIZED_VIEW:
            type_str = 'MATERIALIZED VIEW'
        else:
            type_str = 'TABLE'

        buf.write('DROP %s ' % type_str)
        if if_exists:
            buf.write('IF EXISTS ')
        if schema_name is not None:
            buf.write('%s.%s.`%s`' % (project_name, schema_name, table_name))
        else:
            buf.write('%s.`%s`' % (project_name, table_name))

        return buf.getvalue()

    @utils.with_wait_argument
    def delete(self, table_name, if_exists=False, async_=False, hints=None):
        if isinstance(table_name, Table):
            table_name = table_name.name

        sql = self._gen_delete_table_sql(table_name, if_exists=if_exists)

        del self[table_name]  # release table in cache

        from .tasks import SQLTask
        task = SQLTask(name='SQLDropTableTask', query=sql)

        hints = hints or {}
        hints['odps.sql.submit.mode'] = ''
        schema_name = self._get_schema_name()
        if schema_name is not None:
            hints["odps.sql.allow.namespace.schema"] = "true"
            hints["odps.namespace.schema"] = "true"
        task.update_sql_settings(hints)
        instance = self._parent.project.instances.create(task=task)

        if not async_:
            instance.wait_for_success()
        else:
            return instance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import OrderedDict
from datetime import datetime, date
from decimal import Decimal

try:
    import numpy as np
except ImportError:
    np = None
try:
    import pandas as pd
except ImportError:
    pd = None

from ... import types
from ....models import TableSchema
from ....compat import six

_np_to_df_types = dict()
_df_to_np_types = dict()


def is_na_func(val):
    try:
        func = getattr(pd, "isna", np.isnan)
        ret = func(val)
        return bool(ret)
    except (TypeError, ValueError):
        return False
    except AttributeError:
        pass

    try:
        return val is None or math.isnan(val)
    except TypeError:
        return False


if np is not None:
    _np_int_types = list(map(np.dtype, [np.int_, np.int8, np.int16, np.int32, np.int64]))
    _np_float_types = list(map(np.dtype, [np.float_, np.float32, np.float64]))

    for np_type in _np_int_types + _np_float_types:
        _np_to_df_types[np_type] = getattr(types, str(np_type))

    _np_to_df_types[np.dtype(np.bool_)] = types.boolean
    _np_to_df_types[np.dtype(np.str_)] = types.string

    _df_to_np_types = dict((v, k) for k, v in six.iteritems(_np_to_df_types))


def np_type_to_df_type(dtype, arr=None, unknown_as_string=False, name=None):
    if dtype in _np_to_df_types:
        return _np_to_df_types[dtype]
    if hasattr(pd, "StringDtype") and isinstance(dtype, pd.StringDtype):
        return types.string

    name = ', field: ' + name if name else ''
    if arr is None or len(arr) == 0:
        if dtype == np.dtype(object) and unknown_as_string:
            return types.string
        raise TypeError('Unknown dtype: %s%s' % (dtype, name))

    for it in arr:
        if it is None or is_na_func(it):
            continue
        if isinstance(it, six.string_types):
            return types.string
        elif isinstance(it, six.integer_types):
            return types.int64
        elif isinstance(it, float):
            return types.float64
        elif isinstance(it, datetime):
            return types.datetime
        elif isinstance(it, Decimal):
            return types.decimal
        elif unknown_as_string:  # not inferred
            return types.string
        else:
            raise TypeError('Unknown dtype: %s%s' % (dtype, name))

    if unknown_as_string:
        return types.string
    raise TypeError('Unknown dtype: %s' % dtype)


def pd_to_df_schema(pd_df, unknown_as_string=False, as_type=None):
    if pd_df.index.name is not None:
        pd_df.reset_index(inplace=True)

    dtypes = pd_df.dtypes
    names = list(pd_df.columns.values)

    df_types = []
    for i in range(len(dtypes)):
        arr = pd_df.iloc[:, i]
        if as_type and names[i] in as_type:
            df_types.append(as_type[names[i]])
            continue
        df_types.append(np_type_to_df_type(dtypes.iloc[i], arr,
                                           unknown_as_string=unknown_as_string,
                                           name=names[i]))

    return TableSchema.from_lists(names, df_types)


def df_type_to_np_type(df_type):
    if df_type in _df_to_np_types:
        return _df_to_np_types[df_type]

    if df_type == types.datetime:
        return np.dtype(datetime)
    elif df_type == types.date:
        return np.dtype(date)
    elif df_type == types.timestamp:
        return np.datetime64(0, 'ns').dtype
    elif df_type == types.decimal:
        return np.dtype(Decimal)
    elif df_type == types.binary:
        return np.dtype('bytes')
    else:
        raise TypeError('Unknown DataFrame dtype: %s' % df_type)


def df_schema_to_pd_dtypes(df_schema):
    names = [col.name for col in df_schema.columns]
    dtypes = [df_type_to_np_type(col.type) for col in df_schema.types]

    import pandas as pd
    return pd.Series(dtypes, index=names)


def cast_composite_sequence(pd_seq, df_type):
    def _cast(v, df_type):
        if v is None:
            return None
        elif isinstance(df_type, types.Primitive):
            return df_type_to_np_type(df_type).type(v)
        elif isinstance(df_type, types.List) and isinstance(v, list):
            return [_cast(it, df_type.value_type) for it in v]
        elif isinstance(df_type, types.Dict) and isinstance(v, dict):
            return OrderedDict((_cast(it_key, df_type.key_type), _cast(it_value, df_type.value_type))
                               for it_key, it_value in six.iteritems(v))
        else:
            raise TypeError('Cannot convert value %r into dtype %s' % (v, df_type))

    return pd_seq.map(lambda v: _cast(v, df_type))



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *

cdef class Checksum:

    cdef uint32_t _checksum
    cdef int use_c

    cdef void c_update_bool(self, bint val) nogil
    cdef void c_update_int(self, int32_t val) nogil
    cdef void c_update_long(self, int64_t val) nogil
    cdef void c_update_float(self, float val) nogil
    cdef void c_update_double(self, double val) nogil
    cdef void c_update(self, const char *ptr, size_t length) nogil
    cdef uint32_t c_getvalue(self) nogil
    cdef uint32_t c_setvalue(self, uint32_t val) nogil
    cdef void c_reset(self) nogil

    cpdef update_bool(self, bint val)
    cpdef update_int(self, int32_t val)
    cpdef update_long(self, int64_t val)
    cpdef update_float(self, float val)
    cpdef update_double(self, double val)
    cpdef update(self, bytes b)
    cpdef uint32_t getvalue(self)
    cpdef uint32_t setvalue(self, uint32_t val)
    cpdef reset(self)


# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from ..expr.exporters import get_input_field_names
from ..expr.mixin import ml_collection_mixin
from ...compat import Enum
from ...df import DataFrame


class NetworkFieldRole(Enum):
    FROM_VERTEX = 'FROM_VERTEX'
    TO_VERTEX = 'TO_VERTEX'
    VERTEX_ID = 'VERTEX_ID'
    VERTEX_LABEL = 'VERTEX_LABEL'
    FROM_VERTEX_LABEL = 'FROM_VERTEX_LABEL'
    TO_VERTEX_LABEL = 'TO_VERTEX_LABEL'
    VERTEX_WEIGHT = 'VERTEX_WEIGHT'
    EDGE_WEIGHT = 'EDGE_WEIGHT'


@ml_collection_mixin
class NetworkDFMixIn(object):
    __slots__ = ()

    field_role_enum = NetworkFieldRole
    non_feature_roles = set((NetworkFieldRole.FROM_VERTEX, NetworkFieldRole.TO_VERTEX, NetworkFieldRole.VERTEX_ID,
                             NetworkFieldRole.FROM_VERTEX_LABEL, NetworkFieldRole.TO_VERTEX_LABEL))

"""
Common graph exporters
"""
get_from_vertex_column = partial(get_input_field_names, field_role=NetworkFieldRole.FROM_VERTEX)
get_to_vertex_column = partial(get_input_field_names, field_role=NetworkFieldRole.TO_VERTEX)
get_vertex_id_column = partial(get_input_field_names, field_role=NetworkFieldRole.VERTEX_ID)
get_vertex_label_column = partial(get_input_field_names, field_role=NetworkFieldRole.VERTEX_LABEL)
get_from_vertex_label_column = partial(get_input_field_names, field_role=NetworkFieldRole.FROM_VERTEX_LABEL)
get_to_vertex_label_column = partial(get_input_field_names, field_role=NetworkFieldRole.TO_VERTEX_LABEL)
get_vertex_weight_column = partial(get_input_field_names, field_role=NetworkFieldRole.VERTEX_WEIGHT)
get_edge_weight_column = partial(get_input_field_names, field_role=NetworkFieldRole.EDGE_WEIGHT)


def graph_has_vertex_weight(expr, input_name):
    vert_col = get_vertex_weight_column(expr, 'vertexWeightCol', input_name)
    return 'true' if vert_col else 'false'


def graph_has_edge_weight(expr, input_name):
    edge_col = get_edge_weight_column(expr, 'edgeWeightCol', input_name)
    return 'true' if edge_col else 'false'


"""
Metrics
"""


def get_modularity_result(expr, odps):
    return DataFrame(odps.get_table(expr.tables[0])).execute()[0][0]



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""

import struct

from . import errors
from . import wire_format
from . import output_stream


class Encoder(object):
    """Encodes logical protocol buffer fields to the wire format."""

    def __init__(self):
        self._stream = output_stream.OutputStream()

    def tostring(self):
        """Returns all values encoded in this object as a string."""
        return self._stream.tostring()

    def __len__(self):
        return len(self._stream)

    def append_tag(self, field_number, wire_type):
        """Appends a tag containing field number and wire type information."""
        self._stream.append_var_uint32(wire_format.pack_tag(field_number, wire_type))

    def append_uint32(self, unsigned_value):
        """Appends an unsigned 32-bit integer to our buffer, varint-encoded."""
        self._stream.append_var_uint32(unsigned_value)

    def append_uint64(self, unsigned_value):
        """Appends an unsigned 64-bit integer to our buffer, varint-encoded."""
        self._stream.append_var_uint64(unsigned_value)

    def append_sint32(self, value):
        """Appends a 32-bit integer to our buffer, zigzag-encoded and then
        varint-encoded.
        """
        zigzag_value = wire_format.zig_zag_encode(value)
        self._stream.append_var_uint32(zigzag_value)

    def append_sint64(self, value):
        """Appends a 64-bit integer to our buffer, zigzag-encoded and then
        varint-encoded.
        """
        zigzag_value = wire_format.zig_zag_encode(value)
        self._stream.append_var_uint64(zigzag_value)

    def append_fixed32(self, unsigned_value):
        """Appends an unsigned 32-bit integer to our buffer, in little-endian
        byte-order.
        """
        self._stream.append_little_endian32(unsigned_value)

    def append_fixed64(self, unsigned_value):
        """Appends an unsigned 64-bit integer to our buffer, in little-endian
        byte-order.
        """
        self._stream.append_little_endian64(unsigned_value)

    def append_sfixed32(self, value):
        """Appends a signed 32-bit integer to our buffer, in little-endian
        byte-order.
        """
        sign = (value & 0x80000000) and -1 or 0
        if value >> 32 != sign:
            raise errors.EncodeError('SFixed32 out of range: %d' % value)
        self._stream.append_little_endian32(value & 0xffffffff)

    def append_sfixed64(self, value):
        """Appends a signed 64-bit integer to our buffer, in little-endian
        byte-order.
        """
        sign = (value & 0x8000000000000000) and -1 or 0
        if value >> 64 != sign:
            raise errors.EncodeError('SFixed64 out of range: %d' % value)
        self._stream.append_little_endian64(value & 0xffffffffffffffff)

    def append_float(self, value):
        """Appends a floating-point number to our buffer."""
        self._stream.append_raw_bytes(struct.pack('f', value))

    def append_double(self, value):
        """Appends a double-precision floating-point number to our buffer."""
        self._stream.append_raw_bytes(struct.pack('d', value))

    def append_bool(self, value):
        """Appends a boolean to our buffer."""
        self._stream.append_varint32(value)

    def append_string(self, value):
        """Appends a length-prefixed string to our buffer, with the
        length varint-encoded.
        """
        self._stream.append_var_uint32(len(value))
        self._stream.append_raw_bytes(value)

    def append_bytes(self, value):
        """Appends a length-prefixed sequence of bytes to our buffer, with the
        length varint-encoded.
        """
        self.append_string(value)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import sys


def get_property(name, args=None):
    environ_name = '_'.join(n.upper() for n in name.split('.'))
    args = args or sys.argv[1:]
    for arg in args:
        if arg.startswith('-D%s=' % arg):
            _, value_part = arg.split('=', 1)
            if value_part.startswith('\"') and value_part.endswith('\"'):
                value_part = value_part[1:-1].replace('""', '"')
            return value_part
    return get_environ(environ_name)


# resolve the mdzz environ values
def get_environ(key, default=None):
    val = os.environ.get(key)
    if val:
        if val.startswith('"'):
            val = val.strip('"')
        return val
    return default


def build_image_name(app_name, with_dist_suffix=True):
    from .config import options

    prefix = options.cupid.image_prefix
    version = options.cupid.image_version
    if with_dist_suffix:
        dist_suffix = "-cp" + "".join(str(x) for x in sys.version_info[:2]) + "-" + platform.machine() + "-cpu"
    else:
        dist_suffix = ""

    if prefix is None:
        if version is None:
            return None
        else:
            return app_name + ':' + version + dist_suffix
    else:
        return prefix + app_name + ':' + version + dist_suffix
